# /usr/bin/env python
# -*- encoding: utf-8 -*-
# möp
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#

import os
import re
import string
import sys
from collections import OrderedDict

from atomhandling import get_atomtypes
from elements import get_radius_from_element
from misc import distance, find_line, flatten, get_overlapped_chunks, remove_partsymbol, shift, vol_tetrahedron

# all upper case for case insensitivity:
alphabet = [i for i in string.ascii_uppercase]

# note: parts    1, 2, 3 are _a, _b, _c
# note: residue number 1, 2, 3 are _1, _2, _3
# number first, then part
# part 2 and class 3 = _3b




def remove_duplicate_bonds(bonds):
    """
    removes duplicates from [(at1, at2, 1.324), (at2, at1, 1.324)]
    """
    new_bonds = []
    pairs = []
    for k in bonds:
        pairs.append((k[0], k[1]))
        if (k[1], k[0]) in pairs:
            continue
        new_bonds.append(k)
    return new_bonds


def format_atom_names(atoms, part='', resinum=''):
    """
    needs a list of atoms ['C1', 'C2', 'O1', ..] with part number and a residue number.
    returns a list with atoms like ['C1_4b', 'C2_4b', 'O1_4b', ..]
    :param atoms:  list of plain atom names
    :param part:  string, part number
    :param resinum: string, residue number
    """
    numpart = ''
    part = str(part)
    resinum = str(resinum)
    if not resinum:
        resinum = ''
    try:
        int(part)
        if int(part) > 0:
            partsymbol = alphabet[int(part) - 1]  # turns part number into a letter
        else:
            print('Warning! Part symbol with non-numeric character detected.')
            partsymbol = ''
    except ValueError:
        partsymbol = ''
    if resinum and partsymbol:
        numpart = '_' + resinum + partsymbol
    if not resinum and partsymbol:
        numpart = '_' + partsymbol
    if resinum and not partsymbol:
        numpart = '_' + resinum
    if not resinum and not partsymbol:
        numpart = ''
    # add the _'num''partsymbol' to each atom to be able to find them in the
    # list file:
    atomnames = [i + numpart for i in atoms]
    return atomnames


class Restraints():
    """
    returns an adjacence matrix for all atoms in the .lst file.
    edge property is the bond length

    needs atoms with numpart like C1_2b
    """

    def __init__(self, frag, gdb):
        self.fragment = frag.lower()
        self.gdb = gdb
        self._atoms = self.gdb.get_atoms(fragment=self.fragment, cartesian=False)
        self._atom_names = [i[0] for i in self._atoms]
        self.atom_types = get_atomtypes(self._atoms)
        self.cart_coords = self.gdb.get_coordinates(self.fragment, cartesian=True)
        self._connectivity_table = self.get_conntable_from_atoms(self.cart_coords, self.atom_types, self._atom_names)
        self.coords_dict = self.get_coords_dict()
        self._G = self.get_adjmatrix()

    def get_coords_dict(self):
        """
        Returns an ordered dictionary with coordinates of the fragment
        """
        coords = OrderedDict({})
        for name, co in zip(self._atom_names, self.cart_coords):
            coords[name] = co
        return coords

    def adjmatrix(self):
        """
        create a distance matrix for the atom coordinates
        """
        import networkx as nx
        G = nx.Graph()
        for i in self._connectivity_table:
            atom1 = i[0]
            atom2 = i[1]
            if atom1 in self._atom_names:
                coord1 = self.coords_dict[atom1]
                coord2 = self.coords_dict[atom2]
                dist = distance(coord1[0], coord1[1], coord1[2],
                                coord2[0], coord2[1], coord2[2])
                G.add_edge(atom1, atom2, weight=dist)
        return G

    def get_conntable_from_atoms(self, cart_coords, atom_types, atom_names, extra_param=0.16):
        """
        returns a connectivity table from the atomic coordinates and the covalence
        radii of the atoms.
        TODO:
        - read FREE command from db to control binding here.
        :param cart_coords: cartesian coordinates of the atoms
        :type cart_coords: list
        :param atom_types: Atomic elements
        :type atom_types: list of strings
        :param atom_names: atom name in the file like C1
        :type atom_names: list of strings
        """
        names = []
        for n, i in enumerate(atom_names, 1):
            names.append([n, i])
        conlist = []
        for co1, typ, n1 in zip(cart_coords, atom_types, names):
            for co2, typ2, n2 in zip(cart_coords, atom_types, names):
                if n1 == n2:
                    continue
                ele1_covrad = get_radius_from_element(typ.capitalize())
                ele2_covrad = get_radius_from_element(typ2.capitalize())
                d = distance(co1[0], co1[1], co1[2], co2[0], co2[1], co2[2], round_out=5)
                # print(d, n1, n2, (ele1.covrad+ele2.covrad)+extra_param, '#', ele1.covrad, ele2.covrad)
                # a bond is defined with less than the sum of the covalence
                # radii plus the extra_param:
                if d <= (ele1_covrad + ele2_covrad) + extra_param and d > (ele1_covrad or ele2_covrad):
                    conlist.append([n2[1], n1[1]])
                    if [n1[1], n2[1]] in conlist:
                        continue
        return (conlist)

    def get_adjmatrix(self):
        return self.adjmatrix()

    def get_12_dfixes(self):
        """
        returns the requested dfixes als list of strings
        """
        dfix = []
        for n, i in self._G.adjacency():
            # print(n, i)
            for i, x in list(i.items()):
                dist = x['weight']  # weight of edge is the atomic distance
                atom1 = n
                atom2 = i
                dfix.append((atom1, atom2, dist))
        return dfix

    def get_formated_12_dfixes(self):
        dfix_formated = []
        dfix_restraints = self.get_12_dfixes()
        dfix_restraints = remove_duplicate_bonds(dfix_restraints)
        for n, i in enumerate(dfix_restraints, 1):  # @UnusedVariable
            dfix_formated.append('DFIX {:<7.4f}{:<4s} {:<4s}\n'.format(i[2],
                                                                       remove_partsymbol(i[0]),
                                                                       remove_partsymbol(i[1])))
        return dfix_formated

    def get_neighbors(self, atoms):
        """
        returns the neighbors of the fragment atoms
        """
        from networkx import exception
        neighbors = []
        nb = None
        for at in atoms:
            try:
                nb = self._G.neighbors(at)
            except exception.NetworkXError:
                print('Information: Atom "{}" has no neighbours.'.format(at))
                # sys.exit()
            neighbors.append([at, nb])
        return (neighbors)

    def get_next_neighbors(self):
        """
        returns the next-neighbors of the fragments atoms
        """
        nn = []
        nb12 = self.get_neighbors(self._atom_names)
        for i in nb12:
            atom1 = i[0]
            bonded = list(i[1])
            try:
                for n in bonded:
                    pass
            except(KeyError, TypeError):
                print('No atom connections found. DFIX generation not possible!'
                      ' Check your SHELXL listing file.')
                sys.exit()
            for n in bonded:
                nb = [x for x in self._G.neighbors(n)]
                if not nb:
                    continue
                try:
                    nb.remove(atom1)
                except Exception:
                    # print('Atom {0} has no neighbour.'.format(atom1))
                    pass
                for at in nb:  # nb -> neighbors of n
                    nn.append((atom1, at))
        return nn

    def make_13_dist(self, nn):
        """
        return 1,3-distance as [('at1', 'at2', 'distance'), ('at1', 'at2', 'distance'), ...]
        needs next neighbors pairs
        """
        dist_13 = []
        for i in nn:
            atom1 = i[0]
            atom2 = i[1]
            c1 = self.coords_dict[atom1]
            c2 = self.coords_dict[atom2]
            dist_13.append((atom1, atom2, distance(c1[0], c1[1], c1[2],
                                                   c2[0], c2[1], c2[2])))
        return dist_13

    def make_flat_restraints(self):
        """
        searches for rings in the graph G, splits it in 4-member chunks and tests if
        they are flat: volume of tetrahedron of chunk < 0.1 A-3.
        Additionally, the ring adjacent atoms are added and new chunks created.

        returns list of flat chunks.

        >>> from dbfile import ParseDB
        >>> gdb = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> res = Restraints('benzene', gdb)
        >>> sorted(res.make_flat_restraints())
        [['C1', 'C2', 'C3', 'C4'], ['C3', 'C4', 'C5', 'C6']]
        """
        import networkx as nx
        list_of_rings = sorted(nx.cycle_basis(self._G))
        if not list_of_rings:
            return False
        flats = []
        neighbors = []
        newflats = []
        for ring in list_of_rings:
            ring = sorted(ring)
            for atom in ring:
                # lets see if there is a neighboring atom:
                nb = self._G.neighbors(atom)  # [1:]
                for i in nb:
                    if i not in sorted(flatten(list_of_rings)):
                        neighbors.append(i)
            if len(ring) < 4:
                continue  # only proceed if ring is bigger than 3 atoms
            chunks = get_overlapped_chunks(ring, 4)
            for chunk in chunks:
                if self.is_flat(chunk) and chunk not in flats:
                    flats.append(sorted(chunk))
            if not flats:
                return False
            newflats = []
            # check for neighbours and add the to the flat list:
            for chunk in flats:
                newflats.append(chunk)
                for atnum, chunkatom in enumerate(chunk[:]):
                    for nbatom in neighbors:
                        if self.binds_to(nbatom, chunkatom):
                            # add bound atoms near their partners:
                            ch = chunk[:]
                            ch.insert(atnum, nbatom)
                            ch = shift(ch, atnum)
                            H = self._G.subgraph(ch).copy()
                            # Try to delete atoms in the subgraph and test if subgraph divides.
                            # If it not devides, remove the atom unless it is the just added neighbour.
                            for num, i in enumerate(reversed(ch), start=1):
                                # print(ch, nbatom, ch[-num], num, '###')
                                H.remove_node(ch[-num])
                                comp = list(nx.connected_components(H))
                                # check if graph is disconnected now:
                                if len(comp) > 1:
                                    continue
                                else:
                                    # do not delete the just added neighbour:
                                    if ch[-num] in neighbors:
                                        continue
                                    del ch[-num]
                                    break  # finished, go to next flat
                            # only add if it really results in a flat composition:
                            ch.sort()
                            if self.is_flat(ch) and ch not in newflats:
                                newflats.append(ch)
        return newflats

    def binds_to(self, a, b):
        """
        returns True if atom a binds to atoms b
        """
        if [a, b] in self._connectivity_table:
            return True
        else:
            return False

    def is_flat(self, chunk, flat_tresh=0.085):
        # type: (list, float) -> bool
        """
        check if four atoms are flat
        chunk: a list of four atoms coordinates triples
        flat_tresh: Threshold up to whic volume a tetrahedron is flat.
        """
        tetrahedron_atoms = []
        for atom in chunk:
            single_atom_coordinate = self.coords_dict[atom]
            tetrahedron_atoms.append(single_atom_coordinate)
        a, b, c, d = tetrahedron_atoms
        volume = (vol_tetrahedron(a, b, c, d))
        if volume < flat_tresh:
            return True
        else:
            # print('volume of', chunk, 'too big:', volume)
            return False

    def get_formated_flats(self):
        """
        formats the FLAT restraints and removes the part symbol
        """
        flats = self.make_flat_restraints()
        if not flats:
            return []
        flat_format = []
        for i in flats:
            i = [remove_partsymbol(x) for x in i]
            flat_format.append('FLAT {}\n'.format(' '.join(i)))
        return flat_format

    def get_formated_13_dfixes(self):
        nextneighbors = remove_duplicate_bonds(self.get_next_neighbors())
        dfixes_13 = self.make_13_dist(nextneighbors)
        dfix_13_format = []
        for i in dfixes_13:
            dfix_13_format.append('DANG {:<7.4f}{:<4s} {:<4s}\n'.format(i[2],
                                                                        remove_partsymbol(i[0]),
                                                                        remove_partsymbol(i[1])))
        return dfix_13_format


class ListFile():
    def __init__(self, basefilename):
        self._listfile = basefilename + '.lst'
        self._coord_regex = r'^\s+ATOM\s+x\s+y\s+z'
        self._conntable_regex = r'^.*radii and connectivity'
        self._listfile_list = self.read_lst_file()
        self._single_atom = None

    @property
    def listfile_list(self):
        '''
        the current list file as list
        '''
        return self._listfile_list

    def read_lst_file(self):
        '''
        reads the .lst file and returns it as list.
        ['Sn1 -       Distance       Angles\n', 'O1_a      2.0997 (0.0088)\n''...']
        '''
        listfile = []
        try:
            with open(self._listfile, 'r') as l:
                for line in l:
                    listfile.append(line)
        except(IOError):
            print('Unable to read file {}.'.format(self._listfile))
            sys.exit()
        return listfile

    def read_conntable(self):
        """
        reads the connectivity table from self._listfile_list
        returns a list of all bonded atom pairs. Symmetry equivalent atoms
        are filtered out.
        """
        # find the start of the conntable
        start_line = find_line(self._listfile_list, self._conntable_regex)
        if start_line:
            for num, line in enumerate(self._listfile_list[start_line:]):
                line = line.split()
                # find the end of covalent radii
                if not line:
                    start_line = num + start_line + 1
                    break
        connpairs = []
        connections_list = []
        for i in self._listfile_list[start_line:]:
            line = i.split()
            if not line:
                break
            if 'found' in line:
                continue
            connections_list.append(i.strip(os.linesep).replace('-', '').split())
        for i in connections_list:
            atom = i.pop(0)
            for connected_atom in i:
                atom = atom.upper()
                connected_atom = connected_atom.upper()
                connpairs.append((atom, connected_atom))
        return connpairs

    def coordinates(self):
        """
        reads all atom coordinates of the lst-file
        returns a dictionary with {'atom' : ['x', 'y', 'z']}
        """
        atom_coordinates = {}
        start_line = int(find_line(self._listfile_list, self._coord_regex)) + 2
        for line in self._listfile_list[start_line:]:
            line = line.split()
            try:
                line[0]
            except IndexError:
                break
            xyz = line[1:4]
            try:
                xyz = [float(i) for i in xyz]
            except ValueError:
                print('No atoms found in .lst file!')
                sys.exit(0)
            atom = {str(line[0]).upper(): xyz}
            atom_coordinates.update(atom)
        return atom_coordinates

    def get_cell(self):
        """
        Returns the unit cell parameters from the list file as list:
        ['a', 'b', 'c', 'alpha', 'beta', 'gamma']
        """
        cell = False
        for num, line in enumerate(self._listfile_list):
            if line.startswith(' CELL'):
                cell = line.split()[2:]
                try:
                    cell = [float(i) for i in cell]
                except ValueError as e:
                    print('{} \nbad cell parameters in line {} in the list file.'.format(e, num + 1))
                    sys.exit()
                break
        if not cell:
            print('Unable to find unit cell parameters in the .lst file.')
            sys.exit()
        return cell

    @property
    def get_lst_cell_parameters(self):
        """
        Returns the unit cell parameters from the list file as list:
        ['a', 'b', 'c', 'alpha', 'beta', 'gamma']
        """
        return self.get_cell()

    @property
    def get_all_coordinates(self):
        """
        return all atom coordinates as property
        {'atom' : ['x', 'y', 'z'], 'atom2' : [...]}
        """
        return self.coordinates()

    def get_single_coordinate(self, atom):
        """
        return the coordinates of a single atom as ['x', 'y', 'z']
        """
        coord = self.coordinates()
        try:
            return coord[atom]
        except(KeyError):
            return None

    def get_afix_number_of_CF3(self):
        """
        returns the afix number of the atom where SHELXL prints the
        difference density at 15 degree intervals
        """
        regex_atom = r"^\sDifference\selectron\sdensity.*at\s15\sdegree"
        line = find_line(self._listfile_list, regex_atom)
        if not line:
            return False
        at1 = self._listfile_list[line].split('.')[0].split()
        return at1[-1]

    def get_bondvector(self, atom=None):
        """
        get the bond vector in terms of atom names around which SHELXL
        calculates the difference density in 15 degree interval
        Y-Z-F1/F2/F3
        at1 = Y
        at2 = Z
        """
        import networkx as nx
        regex = r'^.*is clockwise looking down'
        line = find_line(self.read_lst_file(), regex)
        if not line and atom:
            conn = self.read_conntable()
            G = nx.Graph(conn)
            nb = G[atom]
            for i in list(nb):
                if i[0] == 'C':
                    return (i, atom)
        if not line:
            return '', ''
        at1 = self._listfile_list[line].split('.')[0].split()[-3]
        at2 = self._listfile_list[line].split('.')[0].split()[-1]
        return at1, at2

    def get_difference_density(self, averaged=False):
        """
        returns the difference density values in 15 degree interval
        if averaged is True, returns the averaged values.
        :param averaged: enable averaged values
        :type averaged: boolean
        """
        regex_atom = r'^.*is clockwise looking down'
        line = find_line(self._listfile_list, regex_atom)
        if not line:
            return False
        if not averaged:
            nums = self._listfile_list[line + 1].split()
        else:
            nums = self._listfile_list[line + 3].split()[4:]
        return [int(i) for i in nums]

    def get_degree_of_highest_peak(self):
        """
        returns the position in degree of the highest peak in the
        difference density. This point is assumed as an atom position.
        """
        dens = self.get_difference_density(averaged=True)
        maximum = dens.index(max(dens)) + 1
        return maximum * 15


class Lst_Deviations():
    """
    reads the deviations of the fitted group from the lst-file
    """

    def __init__(self, lst_file):
        self._lst_file = lst_file
        self._dev = self.find_deviations()

    def find_deviations(self):
        """
        parses the deviations of the fitted group and
        returns a dictionary with the results:
        {'C3': '11.773', 'C2': '7.667', 'C1': '8.761', 'C4': '5.700'}
        """
        regex = re.compile(r'^\s+\*\*\s+Atom.*deviates\sby')
        deviations = {}
        for line in self._lst_file:
            if regex.match(line):
                line = line.split()
                deviations[line[2]] = line[5]
        return deviations

    def print_LS_fit_deviations(self):
        """
        pretty output of the LS-fit deviations
        """
        if self._dev:
            print('\n Fragment fit might have failed!')
            print(' Deviations on fitting group:')
            for i in self._dev:
                print(' {:<4}: {:>5} A'.format(i.strip(' \n\r'), self._dev[i][:4]))


if __name__ == '__main__':
    pass



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import re
import sys

import misc


class DSRParser(object):
    """
    Handles the parsing of the DSR command. Run this class only once in a DSR run.
    """

    def __init__(self, reslist):
        """
        :param reslist: list of strings of .res file
        """
        self.reslist = reslist
        self._dsr_regex = r'^rem\s{1,5}DSR\s{1,5}.*'
        self.dsr_line_number = 0  # Will be set to real value by find_dsr_command()
        self.multiline = False
        self.dsr_command_list = self.find_dsr_command()
        # Indicates if RESI is on of off in dsr command. Turns True if activated:
        # class and number are defined in self.resi
        self._resiflag = False
        self.dsr_dict = self.parse_dsr_line()
        self.remove_dsr_command()

    def __str__(self):
        return ' '.join(self.dsr_command_list)

    def find_dsr_command(self):
        """
        find the lines with a DSR command entry and return its line number as
        default or the text string when line is set to True
        """
        hklf_endline = misc.find_line(self.reslist, r'^HKLF\s+[1-6]')
        indexnum = misc.find_multi_lines(self.reslist, self._dsr_regex)
        # This is critical, DSR does not run without finding self._dsr_regex:
        try:
            line_number = int(indexnum[0])
        except IndexError:
            print('*** No proper DSR command found! '
                  'Have you really saved your .res file? ***\n')
            sys.exit()
        if int(line_number) > int(hklf_endline):
            print('*** A DSR command after HKLF is not allowed! '
                  'Check line {} ***'.format(line_number + 1))
            sys.exit()
        if len(indexnum) > 1:
            print('*** Only one DSR command at once is allowed! ***')
            sys.exit(-1)
        dsr_str = str(self.reslist[line_number])  # .encode('ascii')
        if misc.multiline_test(dsr_str):
            self.multiline = True
            # in case of a multiline command, strip the '=' and the newline
            dsr_str = dsr_str.rstrip('\n\r= ')
            dsr_str = re.sub(' +', ' ', dsr_str)  # Why did I do this? At least it does no harm.
            dsr_str = dsr_str + self.reslist[line_number + 1]
        self.dsr_line_number = line_number
        return dsr_str.upper().split()

    @property
    def unit_line(self):
        """
        Returns the line number of the UNIT instruction in the SHELX file.
        :return: line number of UNIT
        """
        unitline = misc.find_line(self.reslist, r'^UNIT\s+\d')
        return unitline

    @property
    def hklf_line(self):
        """
        Returns the line number of the HKLF instruction in the res file.
        :return lin number of HKLF
        """
        hklfline = misc.find_line(self.reslist, r'^HKLF\s+\d')
        return hklfline

    def remove_dsr_command(self):
        """
        Removes the dsr command by adding REM in front of the command.
        """
        txt = ' '.join(self.dsr_command_list)
        # wrap the line after 75 chars:
        # The ! prevents weired code highlighting in ShelXle:
        # dsrlines = textwrap.wrap(txt, 75, initial_indent='REM ', subsequent_indent='REM !')
        # dsrlines = '\n'.join(dsrlines)
        # dsrlines += '\n'
        self.reslist[self.dsr_line_number] = ''  # delete old line
        if self.multiline:
            self.reslist[self.dsr_line_number + 1] = ''  # delete second old line
        # self.reslist[self.dsr_line_number] = dsrlines

    def find_commands(self, command):
        """
        returns the value of the input string argument as string
        :param command: string
        """
        if command in self.dsr_command_list:
            try:
                command_index = self.dsr_command_list.index(command)  # find the index of the command
                command_argument = self.dsr_command_list[command_index + 1]  # get the command value
            except IndexError:
                command_argument = ''
        else:
            command_argument = ''
        return command_argument

    def find_atoms(self, start, *stop):
        """
        Returns the parameter between a single start argument
        and one or several stop arguments as list of strings. This can be atoms or
        other Parameters

        ----------
        :type start: str
        :type stop: str
        :rtype: list
        """
        atindex = 0
        try:
            atindex = self.dsr_command_list.index(start)
        except ValueError:
            if start == 'WITH':
                print('*** No source atoms given! ***')
                sys.exit()
            if start == 'ON':
                print('*** No target atoms given! ***')
                sys.exit()
        atoms = []
        for i in self.dsr_command_list[atindex + 1:]:  # start at the position of the first atom
            atoms.append(i.upper())
            if i in stop:  # stop if stop conditions met
                atoms.pop()  # erase the stop condition from the list
                break
        return atoms

    def minimal_requirements(self):
        """
        Checks if minimal requirements of the dsr command are met.
        E.g. the WITH and ON command
        """
        check = ('WITH', 'ON')
        for i in check:
            try:
                if i not in self.dsr_command_list:
                    raise Exception
            except:
                print('\n*** No "WITH" or "ON" statement in the dsr command line found! ***')
                sys.exit()

    def parse_dsr_line(self):
        """
        returns the different parameters from the dsr command as dict
        It needs find_commands() and find_atoms() to parse the line.

        >>> reslist = [r'rem dsr put oc(cf3)3 with o1 c1 c2 c3 c4 on O1_3 c1_3 q6 Q4 q7 resi cf3 =', r'  PART 2 occ -31']
        >>> dsrp = DSRParser(reslist)
        >>> dic = dsrp.all
        >>> l = sorted(dic)
        >>> for i in l:
        ...     print('{}: '.format(i), dic[i])
        command:  PUT
        dfix:  False
        fragment:  OC(CF3)3
        occupancy:  -31
        part:  2
        resi:  ['CF3']
        source:  ['O1', 'C1', 'C2', 'C3', 'C4']
        split:  False
        target:  ['O1_3', 'C1_3', 'Q6', 'Q4', 'Q7']

        >>> dsrp = DSRParser(['rem dsr put ch2Cl2 with C1 cL2 cl1 on C1 C2 Cl3 ParT 1 reSi ocC 21'])
        >>> dic = dsrp.all
        >>> l = sorted(dic)
        >>> for i in l:
        ...     print('{}: '.format(i), dic[i])
        command:  PUT
        dfix:  False
        fragment:  CH2CL2
        occupancy:  21
        part:  1
        resi:  []
        source:  ['C1', 'CL2', 'CL1']
        split:  False
        target:  ['C1', 'C2', 'CL3']

        >>> dsrp = DSRParser(['rem dsr put ch2Cl2 with C1 cL2 cl1 on C1 C2 Cl3 ParT 1  ocC 21'])
        >>> dic = dsrp.all
        >>> l = sorted(dic)
        >>> for i in l:
        ...     print('{}: '.format(i), dic[i])
        command:  PUT
        dfix:  False
        fragment:  CH2CL2
        occupancy:  21
        part:  1
        resi:  False
        source:  ['C1', 'CL2', 'CL1']
        split:  False
        target:  ['C1', 'C2', 'CL3']
        """
        source = None
        if self.cf3_active:
            splitatom = False
            if 'SPLIT' in self.dsr_command_list:
                splitatom = True
        else:
            splitatom = False
            self.minimal_requirements()
        # syntax:
        # rem dsr put|add|replace fragment with source on target part xx AFIX 17x occ occupancy
        command = self.dsr_command_list[2]
        # get the fragment:
        fragment = self.dsr_command_list[3]
        if fragment in ['CF3', 'CF9'] and 'SPLIT' in self.dsr_command_list:
            print('*** Illegal combination of CF3 or CF9 with SPLIT! Only CF6 with SPLIT is alowed! ***')
            sys.exit()
        # make sure the command is correct:
        command_list = ('PUT', 'REPLACE', 'ADD')
        if command not in command_list:
            print('*** No proper put/replace string found in DSR command line! ***')  # , self._dsr_list
            sys.exit()
        for com in command_list:
            if com in self.dsr_command_list[3:]:
                print('*** PUT or REPLACE must occure once and only directly after REM DSR ***')
                sys.exit()
        # Source and target atoms:
        # In parenteses are one start und one to multiple stop conditions:
        if not self.cf3_active:
            # we need no soure atoms for cf3 groups
            source = self.find_atoms('WITH', 'ON')
        target = self.find_atoms('ON', 'PART', 'OCC', 'RESI', 'DFIX', 'SPLIT', '')
        if 'RESI' in self.dsr_command_list:
            residue = self.find_atoms('RESI', 'PART', 'OCC', 'RESI', 'DFIX', '')
            # RESI is in dsr_list but no residue returns -> only RESI in command line:
            if not residue:
                residue = []  # No residue defined, but RESI is active
            self._resiflag = True  # means residue is active
        else:
            residue = False
            self._resiflag = False  # residue completely turned off
        dfix = False
        if 'DFIX' in self.dsr_command_list:
            dfix = True
        part = self.find_commands('PART')
        if part:
            try:
                float(part)
            except ValueError:
                print('*** Part without numerical value supplied. '
                      'Please give a part number after PART in the DSR command. ***')
                sys.exit(False)
            if float(part) > 999:
                print('*** only 999 parts allowed in SHELXL! ***')
                sys.exit(False)
            try:
                if len(part) > 4:
                    print('*** Illegal part number supplied: {}. Please give just one digit '
                          '(positive or negative) in the DSR command ***'.format(part))
                    sys.exit(False)
            except TypeError:
                # no part specified
                pass
        occupancy = self.find_commands('OCC')
        badocc_message = '*** Occupancy without numerical value supplied. Please define occupancy value after OCC ***'
        badocc_status = False
        if occupancy:
            num = occupancy.split('.')
            fvar = abs(int(num[0])) // 10
            try:
                if float(fvar) > 99:
                    print('*** Only 99 free variables allowed in SHELXL! ***')
                    sys.exit()
            except ValueError:
                badocc_status = True
        if 'OCC' in self.dsr_command_list and not occupancy:
            badocc_status = True
        if badocc_status:
            print(badocc_message)
            sys.exit()
        dsr_dict = {
            'command'  : str(command),
            'fragment' : str(fragment),
            'source'   : source,
            'target'   : target,
            'part'     : part,
            'dfix'     : dfix,
            'occupancy': occupancy,
            'resi'     : residue,
            'split'    : splitatom
        }
        return dsr_dict

    @property
    def cf3_active(self):
        """
        Does it have the CF3-Group keyword?
        :return: True/False
        """
        cf = False
        if self.dsr_command_list[3] in ['CF3', 'CF6', 'CF9']:
            cf = True
        return cf

    @property
    def fragment(self):
        """
        database fragment name
        """
        return self.dsr_dict['fragment'].lower()

    @property
    def occupancy(self):
        """
        occupancy of the fragment
        """
        return self.dsr_dict['occupancy']

    @occupancy.setter
    def occupancy(self, value):
        """
        occupancy of the fragment
        """
        self.dsr_dict['occupancy'] = value

    @property
    def command(self):
        return self.dsr_dict['command']

    @property
    def part(self):
        return self.dsr_dict['part']

    @property
    def source(self):
        return self.dsr_dict['source']

    @property
    def target(self):
        return self.dsr_dict['target']

    @property
    def split(self):
        return self.dsr_dict['split']

    @split.setter
    def split(self, value):
        self.dsr_dict['split'] = value

    @property
    def resi(self):
        """
        resi: empty list, resi class, number or resi class and number
        :rtype: list
        """
        return self.dsr_dict['resi']

    @property
    def resiflag(self):
        """
        defines if residue is on or off
        :rtype: bool
        """
        return self._resiflag

    @resiflag.setter
    def resiflag(self, value):
        """
        Setter for resiflag
        :type value: bool
        """
        self._resiflag = value

    @property
    def dfix(self):
        """
        dfix: bool True/False
        """
        return self.dsr_dict['dfix']

    @property
    def all(self):
        """
        Returns the complete dsr command dictionary
        :return: dict
        """
        return self.dsr_dict


if __name__ == '__main__':

    import doctest

    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))
    else:
        print('{} of {} tests failed'.format(failed, attempted))



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import hashlib
import os
import shutil
import sys
import tarfile
import tempfile

from version import VERSION

urlprefix = "https://dkratzert.de/files/dsr"


from urllib.request import FancyURLopener


class MyOpener(FancyURLopener):
    """
    Sets the user agent of the urllib http request.
    """
    version = 'DSR cmdline {}'.format(VERSION)


myurlopen = MyOpener()


def get_current_dsr_version(silent=False):
    """
    determines the current version of DSR on the web server

    >>> get_current_dsr_version()
    '202'

    Returns
    -------
    version number
    :type: str
    """
    import socket
    socket.setdefaulttimeout(3)
    FancyURLopener.version = "DSR-updater {}".format(VERSION)
    try:
        response = myurlopen.open('{}/version.txt'.format(urlprefix))
    except IOError:
        if not silent:
            print("*** Unable to connect to update server. No Update possible. ***")
        return 0
    version = response.readline().decode('ascii').strip()
    return version


def is_update_needed(silent=False):
    # type: (bool) -> bool
    """
    Decides if an update of DSR is needed.
    :return: bool
    >>> is_update_needed()
    False
    """
    try:
        version = int(get_current_dsr_version(silent))
    except ValueError:
        print('*** Could not get version number from update server. ***')
        return False
    if int(VERSION) < int(version):
        return True
    else:
        return False


def update_dsr(force=False, version=None):
    """
    Updates the running DSR to the current version on the web server.
    #>>> update_dsr(force=True)
    #True
    """
    if not version:
        version = get_current_dsr_version()
    if not version:
        print('*** Could not get current version from server. ***')
    if force:
        status = get_update_package(version)
        if status:
            print('*** Finished updating to version {} ***'.format(version))
            return True
        else:
            return False
    if int(VERSION) < int(version):
        print('*** Current available version of DSR is {} you have installed version {}. '
              'Performing upate ... ***'.format(version, VERSION))
        status = get_update_package(version)
        if status:
            print('*** Finished updating to version {} ***'.format(version))
            return True
        else:
            print('*** Could not update DSR. ***')
            return False
    if (int(VERSION) >= int(version)) and int(version) > 0:
        print('*** DSR is already up to date (version {}) ***'.format(VERSION))
        return False


def get_system():
    """
    Gets the currently running operating system
    :return: system
    >>> get_system()
    'win'
    """
    import platform
    plat = platform.system()
    if plat.upper() == "WINDOWS":
        return "win"
    if plat.upper() == "DARWIN":
        return "mac"
    if plat.upper() == "LINUX":
        return "lin"


def post_update_things(dsrdir):
    """
    Performs some file operations after the update.
    :return: None
    """
    import stat
    plat = get_system()
    upath = os.path.join(dsrdir, "../../dsr")
    try:
        if plat == "win":
            pass
        elif plat == "mac":
            shutil.copy2(os.path.abspath(os.path.join(dsrdir, "../../setup/dsr-mac")), upath)
            st = os.stat(upath)
            os.chmod(upath, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
        elif plat == "lin":
            shutil.copy2(os.path.abspath(os.path.join(dsrdir, "../../setup/dsr-linux")), upath)
            st = os.stat(upath)
            os.chmod(upath, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    except IOError:  # Unable to write in this case
        print('*** Unable to perform update. Please run me with super-user rights, e.g.: "sudo /opt/DSR/dsr -u" ***')


def move_dir(root_src_dir, root_dst_dir, move=True):
    """
    Moves the content of scrdir over destdir and overwrites all files.
    
    :param move: move directory instead of copying it
    :param root_src_dir: source directory
    :param root_dst_dir: target directory
    :return: True/False
    """
    for src_dir, dirs, files in os.walk(root_src_dir):
        dst_dir = src_dir.replace(root_src_dir, root_dst_dir, 1)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for file_ in files:
            src_file = os.path.join(src_dir, file_)
            dst_file = os.path.join(dst_dir, file_)
            if os.path.exists(dst_file):
                os.remove(dst_file)
            if move:
                shutil.move(src_file, dst_dir)
            else:
                shutil.copy2(src_file, dst_dir)
    return True


def get_update_package(version, destdir=None, post=True):
    """
    Downloads the current DSR distribution from the web server and
    updates the files.

    :param post: Defines if post_update_things() should be executed
    :param destdir: Optional destdir instead of DSR_DIR
    :type version: int or string
    :return True/False

    >>> get_update_package('202')
    True
    """
    try:
        dsrdir = os.environ["DSR_DIR"]
    except KeyError:
        print("*** Could not determine the location of DSR. Can not update. ***")
        sys.exit()
    # DSR file:
    response = myurlopen.open('{}/DSR-{}.tar.gz'.format(urlprefix, version))
    try:
        with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
            tmpfile.write(response.read())
    except Exception as e:
        print('*** Update timed out. Try again later. ***')
        print(e)
        return False
    downloaded_sha, tgz_sha = check_checksum(tmpfile, version)
    if downloaded_sha != tgz_sha:
        print('*** Checksum mismatch. Unable to update. If this problem persists, please update manually! ***')
        return False
    else:
        print("*** Checksums matched ***")
    # a temporary directory:
    tmpdir = tempfile.mkdtemp()
    try:
        with tarfile.open(tmpfile.name) as tarobj:
            tarobj.extractall(path=tmpdir)
    except tarfile.ReadError as e:
        print('*** Could not get update from server. If this problem persists, please update manually! ***')
        print('***', e, '***')
        return False
    os.remove(tmpfile.name)
    try:
        if not destdir:
            move_dir(os.path.join(tmpdir, "DSR-{}".format(version)), dsrdir, move=False)
        else:
            move_dir(os.path.join(tmpdir, "DSR-{}".format(version)), destdir, move=False)
    except OSError:
        print('*** Unable to perform update. Please run me with super-user rights, e.g.: "sudo /opt/DSR/dsr -u" ***')
        sys.exit()
    shutil.rmtree(tmpdir, ignore_errors=True)  # cleanup the files
    if post:
        post_update_things(dsrdir)
    return True


def check_checksum(tmpfile, version):
    """
    Creates a SHA512 checksum for the downloaded update package.
    :param tmpfile: dowloadad update package
    :param version: version numberof update
    :return: the two checksums
    """
    # download SHA file:
    response2 = myurlopen.open('{}/DSR-{}-sha512.sha'.format(urlprefix, version))
    downloaded_sha = response2.read()
    # Checksum for program package:
    tgz_sha = sha512_checksum(tmpfile.name)
    return downloaded_sha, tgz_sha


def sha512_checksum(filename, block_size=65536):
    """
    Calculates a SHA512 checksum from a file.

    :param filename:
    :param block_size:
    :return: str

    >>> sha512_checksum("../DSR-207.tar.gz")
    'e8d14033578e0ecce0d6c123a947060f9883fa735d1d3226b4f03f08a7eacecd'
    """
    sha512 = hashlib.sha512()
    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha512.update(block)
    return sha512.hexdigest()


if __name__ == '__main__':
    import doctest

    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))
    else:
        print('{} of {} tests failed'.format(failed, attempted))



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import os
import re
import shutil
import subprocess
import sys

import misc
from constants import sep_line
from misc import find_line, check_file_exist, remove_line
from resfile import ResList



def get_xl_version_string(exe):
    """
    Extracts the version string from a SHELXL executable.
    This is fast and needs no hashes etc.
    :type exe: str
    :param exe: path to SHELXL executable
    """
    try:
        with open(exe, 'rb') as f:
            binary = f.read()
            position = binary.find(b'Version 201')
            if position > 0:
                f.seek(position + 8, 0)  # seek to version string
                version = f.read(6)  # read version string
                return version.decode('ascii')
            else:
                return None
    except(IOError):
        print("Could not determine SHELXL version. DSR might fail to run.")
        return None


class ShelxlRefine():
    """
    A class to do a shelxl refinement. It is only for shelxl 2013 and above!
    The resfilename should be without ending.
    """

    def __init__(self, reslist, resfile_name, find_atoms, options):
        """
        :param reslist: SHELXL .res file as list
        :param resfile_name: SHELXL res file name
        :param find_atoms: FindAtoms() object
        """
        self.options = options
        self._find_atoms = find_atoms
        self._atoms_in_reslist = self._find_atoms.collect_residues()
        self.resfile_name = str(resfile_name)
        self._reslist = reslist
        self._shelx_command = self.find_shelxl_exe()
        self.backup_file = os.path.abspath(str(self.resfile_name + '.dsr-bak'))

        if not self._shelx_command:
            print('\nSHELXL executable not found in system path! No fragment fitting possible.\n')
            print('You can download SHELXL at http://shelx.uni-goettingen.de\n')
            sys.exit()

    def find_shelxl_exe(self):
        """
        returns the appropriate shelxl executable
        """
        names = ['shelxl', 'xl']
        download = 'You can download SHELXL at http://shelx.uni-goettingen.de'
        shx_exe = []
        if self.options.shelxl_ex:
            if not check_file_exist(self.options.shelxl_ex):
                print("SHELXL executable not found! Can not proceed...")
                sys.exit()
            return self.options.shelxl_ex
        for name in names:
            shx_exe.extend(misc.which(name))  # list of shelxl executables in path
            try:
                exe = shx_exe[0]
            except IndexError:
                continue
            version = get_xl_version_string(exe)
            if not version:
                print('Your SHELXL version', exe, 'is too old for this Program')
                print('Please use SHELXL 2013/4 or above!')
                print(download)
                sys.exit()
            version = version.split('/')
            if int(version[0]) < 2013:
                print('Your SHELXL version is too old. Please use SHELXL 2013/4 or above!')
                print(download)
                sys.exit()
            else:
                return exe

    def approx_natoms(self):
        """
        Approximates the number of atoms to set the B array accordingly.
        It is slightly more because q-peaks are also counted as atoms
        """
        number_of_atoms = 0
        for residues in self._atoms_in_reslist:
            number_of_atoms += len(self._atoms_in_reslist[residues])
        barray = number_of_atoms * 7
        if barray <= 3000:
            barray = 3000
        return barray

    def set_refinement_cycles(self, cycles):
        """
        Modifies the number of refinement cycles in the reslist.
        """
        status = check_file_exist(self.resfile_name + '.res')
        if not status:
            print('Error: unable to find res file!')
            sys.exit()
        regex = r'(^L\.S\.\s)|(^CGLS\s)'
        ls_line = misc.find_line(self._reslist, regex)
        ls_list = self._reslist[ls_line].split()
        ls_list[1] = str(cycles)
        self._reslist[ls_line] = '  '.join(ls_list) + '\n'
        return self._reslist

    @property
    def get_refinement_cycles(self):
        """
        Modifies the number of refinement cycles in the reslist.
        """
        status = check_file_exist(self.resfile_name + '.res')
        if not status:
            print('Error: unable to find res file!')
            sys.exit()
        regex = r'(^L\.S\.\s)|(^CGLS\s)'
        ls_line = misc.find_line(self._reslist, regex)
        try:
            cycles = self._reslist[ls_line].split()[1]
        except(IndexError):
            return None
        return cycles

    def remove_acta_card(self):
        """
        removes the ACTA card, bacause we refine with L.S. 0, wich is incompatible!
        """
        regex = r'^ACTA'
        acta_lines = misc.find_multi_lines(self._reslist, regex)
        commands = []
        if acta_lines:
            for i in acta_lines:
                commands.append(self._reslist[i])
                self._reslist[i] = 'REM ' + self._reslist[i]
        return commands

    def restore_acta_card(self, lines):
        """
        restores original acta cards.
        :type lines: list
        :param lines: list of lines with previous acta cards
        """
        if not lines:
            return
        regex = r'^rem\s{1,5}ACTA'
        acta_lines = misc.find_multi_lines(self._reslist, regex)
        if not len(lines) == len(acta_lines):
            return
        if acta_lines:
            for n, i in enumerate(acta_lines):
                self._reslist[i] = lines[n]

    def afix_is_closed(self, line):
        """
        check if last afix before dsr command was closed with afix 0

        - returns False if last AFIX was not closed
        - returns True if last AFIX was closed
        - returns True if no AFIX found at all
        """
        afixes = []
        for num, i in enumerate(self._reslist):
            i = i.upper()
            if num <= line:
                if re.match(r'AFIX', i, re.IGNORECASE):
                    afixes.append(i.split()[1])
        try:
            if str(afixes[-2]) != '0':
                return False  # last afix not closed
            else:
                return True  # last afix is closed
        except(IndexError):
            # in this case, no other afix is present and deletion of afix 9 is save
            return True

    def remove_afix(self, random_num):
        """
        removes the AFIX 9 after refinement.
        note: find_line matches case insensitive
        """
        regex = 'REM ' + random_num
        afix_line = misc.find_line(self._reslist, regex)
        if afix_line:
            remove_line(self._reslist, afix_line, remove=True)
        # only delete afix if last afix before the dsr command was closed
        if self.afix_is_closed(afix_line):
            if afix_line:
                remove_line(self._reslist, afix_line - 1, remove=True)
                afix_line2 = misc.find_line(self._reslist, regex)
                if afix_line2:
                    remove_line(self._reslist, afix_line2, remove=True)
                    remove_line(self._reslist, afix_line2, remove=True)
        else:
            if afix_line:
                self._reslist[afix_line - 1] = 'AFIX 0\n'

    def backup_shx_file(self):
        """
        makes a copy of the res file
        make backup in dsrsaves before every fragment fit.
        name: self.resfile_name-date-time-seconds.res
        """
        import datetime
        now = datetime.datetime.now()
        timestamp = (str(now.year) + '_' + str(now.month) + '_' + str(now.day) + '_' +
                     str(now.hour) + '-' + str(now.minute) + '-' + str(now.second))
        resfile = os.path.abspath(str(self.resfile_name + '.res'))
        bakup_dir = os.path.abspath(os.path.dirname(resfile)) + os.path.sep + os.path.relpath('../dsrsaves')
        try:
            shutil.copyfile(resfile, self.backup_file)
        except IOError:
            print('*** Unable to make backup file from {}. ***'.format(resfile))
            sys.exit()
        if not os.path.exists(bakup_dir):
            try:
                os.makedirs(bakup_dir)
            except(IOError, OSError):
                print('*** Unable to create backup directory {}. ***'.format(bakup_dir))
        try:
            shutil.copyfile(resfile,
                            bakup_dir + os.path.sep + os.path.split(self.resfile_name)[1] + '_' + timestamp + '.res')
        except IOError:
            print('\n*** Unable to make backup file from {} in dsrsaves. ***'.format(resfile))

    def restore_shx_file(self):
        """
        restores filename from backup
        """
        resfile = os.path.abspath(str(self.resfile_name + '.res'))
        try:
            print('*** Restoring previous res file. ***')
            shutil.copyfile(self.backup_file, resfile)
        except IOError:
            print('Unable to restore res file from {}.'.format(self.backup_file))
        try:
            misc.remove_file(self.backup_file)
        except IOError:
            print('Unable to delete backup file {}.'.format(self.backup_file))

    def pretty_shx_output(self, output):
        """
        selectively prints the output from shelx
        """
        wr2 = False
        r1 = False
        gof = False
        for out in output:
            if out.startswith(' +  Copyright(C)'):
                print(' SHELXL {}'.format(' '.join(out.split()[6:8])))
            # wR2
            # These values are always bad after a simple LS fit without any atom movement:
            # if out.startswith(' wR2') and not wr2:
            #    wr2 = True
            #    line = out[:].split()
            #    print(' {}  {} {:>6}'.format(line[0], line[1], line[2][:6]))
            # R1
            # if out.startswith(' R1') and not r1:
            #    r1 = True
            #    line = out[:].split()
            #    print(' {}   {} {:>6}'.format(line[0], line[1], line[2][:6]))
            # GooF
            # if re.match(r'.*GooF.*', out) and not gof:
            #    gof = True
            #    line = out.split()
            #    print(' {} {} {:>5}0'.format(line[0], line[1], line[4][:5]))
            if re.match(r'.*CANNOT RESOLVE (SAME|RIGU|SIMU|DELU)', out):
                print('\nWarning: Are you sure that all atoms are in the correct order?\n')
            if re.match(r'.*CANNOT\s+OPEN\s+FILE.*hkl.*', out):
                print('*** No hkl file found! ***')
                print('*** You need a proper hkl file to use DSR! ***')
                sys.exit()
            if re.match(r'.*\*\* Extinction \(EXTI\) or solvent.*', out):
                continue
            if re.match(r'.*\*\* MERG code changed to 0', out):
                # disable this output
                continue
            if re.match(r'.*\*\* Bond\(s\) to .* ignored', out):
                # disable this output
                continue
            if re.match(r'.*\*\*.*', out):
                print('\n SHELXL says:')
                print(' {}'.format(out.strip('\n\r')))

    def run_shelxl(self):
        """
        This method runs shelxl 2013 on the res file self.resfile_name

        TODO: Do the fit with https://github.com/tjmustard/QuatfitPy
        """
        resfile = self.resfile_name + '.res'
        hklfile = self.resfile_name + '.hkl'
        if not check_file_exist(hklfile):
            print('You need a proper hkl file to use DSR.')
            sys.exit()
        command_line = ['{}'.format(self._shelx_command), "-b{}".format(self.approx_natoms()),
                        '{}'.format(self.resfile_name)]
        self.backup_shx_file()
        print(sep_line)
        print(' Running SHELXL with "{}" and "L.S. 0"'.format(' '.join(command_line)))
        p = subprocess.Popen(command_line,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        (child_stdin, child_stdout_and_stderr) = (p.stdin, p.stdout)
        child_stdin.close()
        # Watch the output for successful termination
        out = child_stdout_and_stderr.readline().decode('ascii')
        output = []
        while out:
            output.append(out)
            out = child_stdout_and_stderr.readline().decode('ascii')
        child_stdout_and_stderr.close()
        # output only the most importand things from shelxl:
        self.pretty_shx_output(output)
        status = check_file_exist(resfile)  # status is False if shelx was unsecessful
        if not status:  # fail
            print(sep_line)
            print('\nError: SHELXL terminated unexpectedly.')
            print('Check for errors in your SHELX input file!\n')
            self.restore_shx_file()
            sys.exit()

    def check_refinement_results(self, list_file):
        """
        Does some checks if the refinement makes sense e.g. if the data to parameter
        ratio is in an acceptable range.
        :param list_file: SHELXL listing file
        :type list_file: list
        """
        is_resfile_there = misc.check_file_exist(self.resfile_name + '.res')
        if is_resfile_there and is_resfile_there == 'zero':
            print('Something failed in SHELXL. Please check your .ins and .lst file!')
            self.restore_shx_file()
            try:
                misc.remove_file(self.backup_file)
            except IOError:
                print('Unable to delete backup file {}.'.format(self.backup_file))
            sys.exit()
        if not is_resfile_there:
            print('Something failed in SHELXL. Please check your .ins and .lst file!')
            self.restore_shx_file()
            try:
                misc.remove_file(self.backup_file)
            except IOError:
                print('Unable to delete backup file {}.'.format(self.backup_file))
            sys.exit()
        regex_final = r' Final Structure Factor Calculation.*\n'
        final_results = find_line(list_file, regex_final)
        # find data and parameters:
        try:
            dataobj = re.search(r'[0-9]+\s+data', list_file[final_results + 4])
            data = float(dataobj.group(0).split()[0])
            parameterobj = re.search(r'[0-9]+\s+parameters', list_file[final_results + 4])
            parameters = float(parameterobj.group(0).split()[0])
            restrobj = re.search(r'[0-9]+\s+restraints', list_file[find_line(list_file, r" GooF = S =.*")])
            restraints = float(restrobj.group(0).split()[0])
        except AttributeError:
            return False
        try:
            data_to_parameter_ratio = data / parameters
            restr_ratio = ((data + restraints) / parameters)
        except ZeroDivisionError:
            return False
        lattline = find_line(list_file, r'^ LATT.*')
        centro = None
        if lattline:
            try:
                latt = int(list_file[lattline].split()[1])
            except ValueError:
                latt = False
            if latt > 0:
                centro = True
            else:
                centro = False
        if centro and data_to_parameter_ratio < 10:
            print('*** Warning! The data/parameter ratio is getting low (ratio = {:.1f})! ***'
                  '\n*** but consider (data+restraints)/parameter = {:.1f} ***'
                  .format(data_to_parameter_ratio, restr_ratio))
        if not centro and data_to_parameter_ratio < 7.5:
            print('*** Warning! The data/parameter ratio is getting low (ratio = {:.1f})! ***'
                  '\n*** but consider (data+restraints)/parameter = {:.1f} ***'
                  .format(data_to_parameter_ratio, restr_ratio))
        try:
            misc.remove_file(self.backup_file)
        except IOError:
            print('Unable to delete backup file {}.'.format(self.backup_file))


if __name__ == '__main__':
    from atomhandling import FindAtoms
    import options

    res_file = 'p21c.res'
    rl = ResList(res_file)
    res_list = rl.get_res_list()
    find_atoms = FindAtoms(res_list)
    # lf = ListFile('p21c')
    # lst_file = lf.read_lst_file()
    options = options.OptionsParser('program_name dsr')
    shx = ShelxlRefine(res_list, 'testfile', find_atoms, options)

    # print(shx.afix_is_closed(110))
    # print(shx.remove_afix())
    # shx.set_refinement_cycles()
    shx.run_shelxl()



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import re
import string
import sys

from atoms import Element, atoms
from constants import atomregex, SHX_CARDS
from misc import find_line, find_multi_lines, atomic_distance



def get_atoms(atlist):
    """
    returns all atoms found in the input as list of lists
    >>> get_atoms(['F8    4    0.349210   0.073474   0.519443 -21.00000   0.03106', 'foo'])
    [['F8', '4', '0.349210', '0.073474', '0.519443']]
    """
    atoms = []
    try:
        atlist[0]
    except IndexError:
        return []
    for i in atlist:
        if atomregex.match(str(i)):  # search atoms
            l = i.split()[:5]  # convert to list and use only first 5 columns
            if l[0].upper() not in SHX_CARDS:  # exclude all non-atom cards
                atoms.append(l)
    return atoms


def get_atomtypes(dbatoms):
    """
    find all atoms in a list of shelxl format atom lines.
    returns a list like ['N', 'C', 'C', 'C'].

    :param dbatoms:  atoms in a database entry
                     [['O1', 3, '-0.01453', '1.66590', '0.10966'],
                      ['C1', 1, '-0.00146', '0.26814', '0.06351'],
                      [...]
                     ]
    :type dbatoms: list
    >>> atoms = [['O1', 3, '-0.01453', '1.66590', '0.10966'],\
                 ['C1', 1, '-0.00146', '0.26814', '0.06351'],\
                 ['Au15', 8, '0.00146', '0.26814', '0.06351']]
    >>> get_atomtypes(atoms)
    ['O', 'C', 'AU']
    >>> atoms = [['O1', 3, '-0.01453', '1.66590', '0.10966'],\
                 ['C1', 1, '-0.00146', '0.26814', '0.06351'],\
                 ['Ax15', 8, '0.00146', '0.26814', '0.06351']]
    >>> get_atomtypes(atoms)
    Traceback (most recent call last):
    ...
    KeyError
    """
    el = Element()
    found = []
    # find lines with atoms and see if they are in the atom list
    for i in dbatoms:
        sfacnum = i[1]
        atom_name = i[0].upper()  # i is the full atom name with number suffix like C1
        try:
            if int(sfacnum) < 0:
                found.append(el.get_element(abs(int(sfacnum))))
                continue
        except:
            pass
        found.append(el.get_atomlabel(atom_name))
    if len(dbatoms) != len(found):  # do we really need this here??
        print("*** One of the Atoms in the database entry is not correct! ***")
        raise KeyError
    return found


def replace_after_fit(rl, reslist, resi, fragment_numberscheme, cell):
    """
    deletes the atoms in replace mode that are near the fragment atoms

    :param rl: Reslist() instance
    :param reslist: .res file list
    :param resi: Resi() instance
    :param fragment_numberscheme: atom names of the fitting fragment
    :param cell: cell parameters

    >>> from resfile import ResList
    >>> from resfile import ResListEdit
    >>> from dsrparse import DSRParser
    >>> from resi import Resi
    >>> res_file = 'tests/p21c.res'
    >>> rl = ResList(res_file)
    >>> reslist = rl.get_res_list()
    >>> find_atoms = FindAtoms(reslist)
    >>> rle = ResListEdit(reslist, find_atoms)
    >>> dsrp = DSRParser(reslist)
    >>> dbhead = ['RESI CF3']
    >>> db_residue_string='CF3'
    >>> resi = Resi(dsrp, db_residue_string, find_atoms)
    No residue number was given. Using residue number 4.
    >>> fragment_numberscheme = ['O1_1', 'C1_1', 'C2_1', 'F1_1', 'F2', 'F3', 'C3', 'F4', 'F5', 'F6', 'C4', 'F7', 'F8', 'F9']
    >>> cell = rle.get_cell()

    #>>> replace_after_fit(rl, reslist, resi, fragment_numberscheme, cell)
    """
    remdist = 1.3
    from resfile import ResListEdit
    find_atoms = FindAtoms(reslist)
    if resi.dsrp.resiflag:
        frag_at = []
        for i in fragment_numberscheme:
            at = i + '_{}'.format(resi.get_resinumber)
            frag_at.append(at)
    else:
        frag_at = fragment_numberscheme
    atoms_to_delete = find_atoms.find_atoms_to_replace(frag_at, cell, remdist)
    if atoms_to_delete:
        print('Replacing following atoms (< {0} A near fragment):\n'.format(remdist),
              ' '.join(sorted(set(atoms_to_delete))))
    else:
        print("No replaceable atoms found.")
    target_lines = set(find_atoms.get_atom_line_numbers(atoms_to_delete))
    rle = ResListEdit(reslist, find_atoms)
    for i in target_lines:
        i = int(i)
        rle.remove_line(i, rem=False, remove=True, frontspace=False)
    rl.write_resfile(reslist, '.res')
    reslist = rl.get_res_list()
    return reslist, find_atoms


class FindAtoms():
    """
    Finds Atoms and creates a data structure
    """

    def __init__(self, reslist):
        """
        :param reslist: SHELXL .res file as list
        """
        self._reslist = reslist
        sfacline = find_multi_lines(self._reslist, r'SFAC\s+[a-zA-Z]+')
        try:
            self.sfac = self._reslist[sfacline[0]].split()[1:]
        except IndexError:
            print("*** No SFAC card found in .res file ***")
            sys.exit()
        self.e2s = Elem_2_Sfac(self.sfac)
        self._residues = self.collect_residues()

    def atoms_as_residues(self):
        """
        returns   residues = { #        0           1            2         3      4      5          6
                                {'0': ['C1', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number],
                                      ['C2', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number]
                                },
                             {'1': ['C1', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number],
                                   []}
                             }
        """
        return self._residues

    @staticmethod
    def is_atom(atomline):
        """
        returns all atoms found in the input as list if they are real atoms

        :param atomline:  'O1    3    0.120080   0.336659   0.494426  11.00000   0.01445 ...'
        >>> FindAtoms.is_atom(atomline = 'O1    3    0.120080   0.336659   0.494426  11.00000   0.01445 ...')
        ['O1', '3', '0.120080', '0.336659', '0.494426']
        >>> FindAtoms.is_atom(atomline = 'O1    0.120080   0.336659   0.494426  11.00000   0.01445 ...')
        []
        >>> FindAtoms.is_atom(atomline = 'O1    0.120080   0.336659   0.494426  11.00000   0.01445 ...')
        []
        """
        if atomregex.match(str(atomline)):  # search atoms
            atom = atomline.split()[:5]  # convert to list and use only first 5 columns
            if atom[0].upper() not in SHX_CARDS:  # exclude all non-atom cards
                return atom
            else:
                return []
        else:
            return []

    @staticmethod
    def get_resi_definition_dict(resi):
        """
        returns the residue number and class of a string like 'RESI TOL 1'
        or 'RESI 1 TOL'
        {'class': 'TOL', 'number': '1'}

        :param resi: ['RESI', 'number', 'class']
        :type resi: list or string

        >>> sorted(list(FindAtoms.get_resi_definition_dict('RESI 1 TOL')))
        ['class', 'number']
        >>> sorted(FindAtoms.get_resi_definition_dict('RESI 1 TOL').values())
        ['1', 'TOL']
        >>> FindAtoms.get_resi_definition_dict('RESI 1 TOL')
        {'class': 'TOL', 'number': '1'}
        >>> FindAtoms.get_resi_definition_dict('RESI')
        {'class': None, 'number': None}
        """
        resi_dict = {
            'class': None,
            'number': None}
        try:
            resi.remove('RESI')
        except AttributeError:
            resi = resi.split()
            resi.remove('RESI')
        resi.sort()
        if len(resi) > 0:
            if str.isalpha(resi[-1][0]):
                resi_dict['class'] = resi.pop()
            if len(resi) > 0 and str.isdigit(resi[0]):
                resi_dict['number'] = resi[0]
                del resi[0]
        return resi_dict

    @staticmethod
    def get_partnumber(partstring):
        """
        get the part number from a string like PART 1 oder PART 2 -21
        Parameters
        ----------
        partstring: string like 'PART 2 -21'

        >>> FindAtoms.get_partnumber(partstring='PART 2 -21')
        2
        """
        partstring = partstring.upper()
        part = partstring.split()
        try:
            partnum = int(part[1])
        except(ValueError, IndexError):
            print('*** Wrong PART definition found! Check your PART instructions ***')
            partnum = 0
        return partnum

    def find_atoms_to_replace(self, frag_atoms, cell, remdist=1.2, only_this=None):
        """
        this method looks around every atom of the fitted fragment and removes
        atoms that are near a certain distance to improve the replace mode

        :param frag_atoms: atoms of the fitting fragment ['C1', '1', 'x', 'y', 'z']
        :param cell: unit cell parameters (list)
        :param remdist: distance below atoms shoud be deleted

        Parameters
        ----------
        only_this: replace only this special atom
        """
        atoms_to_delete = []
        frag_coords = self.get_atomcoordinates(frag_atoms)
        for resinum in self._residues:
            suffix = ''
            if resinum != 0:
                suffix = '_{}'.format(resinum)
            # resinum is the resideue number
            for atom in self._residues[resinum]:
                if atom[0].startswith('Q'):
                    # ignore q peaks:
                    continue
                if only_this and only_this != atom[5]:
                    continue
                # atomlist[4] is the part number
                # replace only in part 0:
                if int(atom[4]) == 0:
                    for name in frag_coords:
                        # name is the atom name
                        if atom[0] + suffix in frag_coords:
                            # do not delete the fitted fragment
                            break
                        at1 = [float(x) for x in frag_coords[name]]
                        at2 = [float(x) for x in atom[1]]
                        resinum1 = self.get_atoms_resinumber(name)
                        resinum2 = self.get_atoms_resinumber(atom[0] + suffix)
                        if at1 == at2 and resinum1 == resinum2:
                            # do not delete atoms on exactly the same position
                            # and same residue
                            break
                        d = atomic_distance(at1, at2, cell, shortest_dist=True)
                        if d < remdist:
                            atoms_to_delete.append(atom[0] + suffix)
        return sorted(atoms_to_delete)

    def collect_residues(self):
        """
        find all atoms and sort them into residues

        residues is a dictionary which includes a dictionary for each residue
        which in turn includes a list of its atoms.

        residues = { {'0': ['C1', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number],
                           ['C2', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number]},
                     {'1': ['C1', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number],
                     []} }
        """
        resi = False
        part = False
        partnum = '0'
        resiclass = None
        residues = {'0': []}
        resinum = 0
        for num, i in enumerate(self._reslist):
            i = i.upper()
            # First collect the residue
            if re.match(r'^RESI\s+0', i) and resi:
                resi = False
                continue
            if i.startswith(('END', 'HKLF')) and resi:
                resi = False
                continue
            if i.startswith('RESI') and not re.match(r'^RESI\s+0', i):
                resi = True
                resinum = self.get_resi_definition_dict(i.split())['number']
                if not resinum:
                    continue
                resiclass = self.get_resi_definition_dict(i.split())['class']
                residues.update({resinum: []})
                continue
            # Now collect the part:
            if re.match(r'^PART\s+0', i) and part:
                part = False
                partnum = 0
                continue
            if i.startswith(('END', 'HKLF')) and part:
                part = False
                continue
            if i.startswith('PART') and not re.match(r'^PART\s+0', i):
                part = True
                partnum = self.get_partnumber(i)
                continue
            #####################
            if resi:
                atom = self.is_atom(i)
                if atom:
                    sfac = atom[1]
                    elem = self.e2s.sfac_2_elem(sfac)
                    residues[resinum].append([atom[0], [float(y) for y in atom[2:5]], num,
                                              resiclass, partnum, elem, sfac])
            else:
                atom = self.is_atom(i)
                resinum = '0'  # all other atoms are residue 0
                if atom:
                    sfac = atom[1]
                    elem = self.e2s.sfac_2_elem(sfac)
                    residues[resinum].append([atom[0], [float(x) for x in atom[2:5]], num,
                                              resiclass, partnum, elem, sfac])
        return residues

    def get_atoms_resiclass(self, atom):
        """
        returns the residue class of a given atom. C1 would be 'None'
        C1_1 would be 'CF3' for example

        :param atom: an atom name with or without residue number like C1 or C1_1
        :type atom: string
        """
        num = self.get_atoms_resinumber(atom)
        atom = atom.split('_')[0]
        residue = self._residues[num]
        if atom in residue[1]:
            #  atom-name  class
            return residue[1][3]

    def get_atoms_resinumber(self, atom):
        """
        returns the residue number of a given atom. C1 would be '0'
        C1_1 would be '1', ...

        :param atom: an atom name with or without residue number like C1 or C1_1
        :type atom: string
        """
        if '_' in atom:
            suffix = atom.split('_')
            resinum = suffix[-1].strip(string.ascii_letters)  # we don't need the part here
            if not resinum:
                resinum = '0'
            if len(resinum) > 4:
                print('*** Invalid residue number in {}'.format(atom))
                sys.exit()
            return str(resinum)
        else:
            return '0'

    def get_atomcoordinates(self, atoms):
        """
        finds an atom regardless if it is in a residue or not.

        rerturns a dictionary {'C1': ['1.123', '0.7456', '3.245']}

        start with digit-> rest auch digit-> resinumber or alias
        start with letter-> rest letter or digit -> residue class

        Negative parts have no letter? part -1 resi 2 is C1_2

        Atom C1 in residue 2 would be addressable with C1_2
        searching for PART is Afaik not neccesary, because inside a residue
        the atom names have to be unique anyway.
        e.g.
        rem DSR put toluene with C1 C2 C3 on C1_2 q2 q3

        :param atoms: list of atoms like ['C1', 'Q2', 'C3_2', ...]
        :type atoms: list
        """
        atom_dict = {}
        for i in atoms:
            num = self.get_atoms_resinumber(i)
            try:
                self._residues[num]
            except KeyError:
                print('*** Atom "{}" not found in res file!! ***'.format(i))
                break
            for x in self._residues[num]:
                if x[0].upper() == i.split('_')[0].upper():
                    single_atom = {i.upper(): x[1]}
                    atom_dict.update(single_atom)
        for i in atoms:
            i = i.upper()
            if i not in list(atom_dict):
                print('\n*** Atom "{}" not found in res file! ***'.format(i))
                sys.exit()
        return atom_dict

    def get_atom_line_numbers(self, atoms):
        """
        returns the line numbers in the res_list of a given atom list
        one atom of self._residues[resinum]:
            ['C12', ['0.471727', '0.649578', '0.232054'], 98]
        returns a list like ['98', 'xx', ..]
        :param atoms: list of atom names
        :type atoms: list
        :return lines: list of integers
        """
        # this is to make sure the atomlines are correct:
        residues = self.collect_residues()
        lines = []
        for at in atoms:
            resinum = self.get_atoms_resinumber(at.upper())
            for x in residues[resinum]:
                atom1 = x[0].upper()
                atom2 = at.split('_')[0].upper()
                if atom1 == atom2:
                    single_atom = x[2]  # x[2] is the line number
                    lines.append(single_atom)
        return lines

    def remove_adjacent_hydrogens(self, atoms, sfac_table, atype='H'):
        """
        if an atom is replaced, its hydrogen atoms are deleted
        this method searches for the first afix behind the atom,
        deletes all lines until AFIX 0, HKLF or the 10th line appears.
        Also Hydrogen atoms outside AFIX are deleted. They are recognized by their
        SFAC number.
        -TODO: create global connectivity table to find atom connections in CF3 CH3 groups
        :param atoms:   list   list of non-Hydrogen atoms
        :param sfac_table: list of atom types ['C', 'H', 'N', ...]
        """
        delcount = []
        if atype == 'H':
            name = "hydrogen"
        if atype == 'F':
            name = "fluorine"
        lines = self.get_atom_line_numbers(atoms)
        try:
            hydrogen_sfac = sfac_table.index(type) + 1
        except ValueError:
            hydrogen_sfac = False
            return
        for i in lines:
            afix = False
            if i == '':
                continue
            if i == '\n':
                continue
            for n in range(1, 11):
                try:
                    line = self._reslist[i + n + 1].upper()
                    atom = line.split()[0].upper()
                except IndexError:
                    continue
                if line.startswith('HKLF'):
                    break  # stop in this case because the file has ended anyway
                if atomregex.match(line) and not afix:
                    # stop if next line is an atom and we are not inside an "AFIX MN"
                    if str(line.split()[1]) == str(hydrogen_sfac):
                        print('Deleted {0} atom {1}'.format(name, atom))
                        delcount.append(atom)
                        self._reslist[i + n] = ''
                        continue
                    else:
                        break
                if line.startswith('AFIX') and afix and line.split()[1] != '0':
                    afix = False
                    # stop also if next "AFIX mn" begins
                    break
                if line.startswith('AFIX') and line.split()[1] != '0':
                    self._reslist[i + n + 1] = ''
                    afix = True  # turn on afix flag if first "AFIX mn" is found
                    continue
                if line.startswith('AFIX') and line.split()[1] == '0':
                    afix = False  # turn of afix flag if afix is closed with "AFIX 0"
                    self._reslist[i + n + 1] = ''
                    continue
                if afix:
                    try:
                        if atom in SHX_CARDS:
                            continue
                        # delete the hydrogen atom
                        self._reslist[i + n + 1] = ''
                        print('Deleted {0} atom {1}'.format(name, atom))
                        delcount.append(atom)
                    except IndexError:
                        continue
        return delcount


def check_source_target(db_source_atoms, res_target_atoms, dbatoms):
    """
    several checks if the atoms in the dsr command line are consistent
    :param db_source_atoms:   ['C1', 'O1', 'C2', ...]
    :param res_target_atoms:  ['C1', 'Q2', 'C3_2', ...]
    :param dbatoms:           [['C1', 1, '-0.00146', '0.26814', '0.06351'],
                               ['C2', 1, '-1.13341', '-0.23247', '-0.90730'], ...]]
    """
    temp = [i[0].upper() for i in dbatoms]
    # check if source and target are of same length:
    nsrc = len(db_source_atoms)
    ntrg = len(res_target_atoms)
    if nsrc != ntrg:
        print('*** Number of source and target atoms/peaks is different!! '
              '({} and {} atoms/peaks) ***'.format(nsrc, ntrg))
        sys.exit()
    # do the source atoms exist at all?:
    for i in db_source_atoms:
        i = i.upper()
        if i not in temp:
            print('\n*** Atom {} not found in database entry! ***'.format(i))
            sys.exit()
    return True


def rename_restraints_atoms(new_atoms, old_atoms, dbhead):
    """
    Returns the restraints with the old atom names replaced by the new
    number scheme fitting to the structures sheme.

    :param new_atoms: ['C1A', 'O1A', 'C2A', ...]
    :param old_atoms: ['C1', 'O1', 'C2', ...]
    :param dbhead:    [SAME F5A F6A F6A F4A F7A F8A F8A F9A F9A F7A\n', 'SIMU O1A > F9A\n',
                'RIGU O1A > F9A\n']
    """
    new = list(reversed(new_atoms))
    headneu = []
    for line in dbhead:
        line = line.split()
        for x, a in enumerate(old_atoms):
            for n, i in enumerate(line):
                if i.upper() == a.upper():
                    line[n] = new[x]
        headneu.append(' '.join(line) + '\n')
    return headneu


def set_final_db_sfac_types(db_atom_types, dbatoms, sfac_table):
    """
    corrects the sfac types of the dbentry according to sfac card of the
    res file
    :param db_atom_types: element names of each atom in the database entry
                          like ['C', 'C', 'N', ... ]
    :type db_atom_types: list
    :param dbatoms: full atoms of the database entry
    :type dbatoms: list
    :param sfac_table: list of scattering factors from SHELXL
    :type sfac_table: list
    """
    e2s = Elem_2_Sfac(sfac_table)
    atype = list(reversed(db_atom_types))
    for line in dbatoms:  # go through db entry
        # replace scattering factor (line[1]) with true one
        line[1] = e2s.elem_2_sfac(atype.pop())
    return dbatoms


class SfacTable():
    """
    Combines the atomtypes of the resfile SFAC and the dbentry to a global
    SFAC table for the new res file.
    """

    def __init__(self, reslist, dbatom_types):
        """

        :param reslist: list
                SHELXL .res file as list
        :param dbatom_types: list
                ['N', 'C', 'C', 'C']
        """
        self._reslist = reslist
        self._db_atom_types = dbatom_types
        self.elements = [x.upper() for x in atoms]

    def set_sfac_table(self):
        """
        sets the new global sfac table in the res file
        """
        sfacline = find_multi_lines(self._reslist, r'SFAC\s+[a-zA-Z]+')  # position of the SFAC card
        unitline = find_line(self._reslist, r'UNIT\s+[0-9]+')  # position of the UNIT card
        try:
            if sfacline[-1] > unitline:
                print('*** SFAC in line {} must be defined before UNIT! ***'.format(sfacline[-1] + 1))
                sys.exit()
        except():
            pass
        unit = []
        explicit_scat = []
        regular_sfac_line_num = False
        sfac = []
        if len(sfacline) == 1:
            # regular SFAC table
            sfac = self._reslist[sfacline[0]].split()[1:]  # SFAC string in the reslist
            sfacline = sfacline[0]
        elif len(sfacline) > 1:
            # first and second type of sfac:
            for i in sfacline:
                if not ''.join(self._reslist[i].split()).isalpha():  # SFAC with scattering factor
                    # in this case the SFAC command defines also a scattering factor:
                    element = self._reslist[i].split()[1]
                    explicit_scat.append(element)  # first SFAC paramter is element
                else:
                    # regular SFAC list
                    sfac.extend(self._reslist[i].split()[1:])
                    regular_sfac_line_num = i
        if regular_sfac_line_num:
            sfacline = regular_sfac_line_num
        if not sfacline:
            print('*** No SFAC card found! Can not proceed ***')
            sys.exit()
        sfac = [x.upper() for x in sfac]
        explicit_scat = [x.upper() for x in explicit_scat]
        for i in self._db_atom_types:  # this is to compare the occurence of element type from resfile and db
            i = i.upper()
            if i not in sfac + explicit_scat:  # all atom types from db not already in sfac
                sfac.append(i)  # get appended to sfac
            if i not in self.elements:
                print('*** Error, atom {} not valid ***'.format(i))
                sys.exit()
        for i in range(1, len(sfac + explicit_scat) + 1):
            i = str(i).upper()
            unit.append(i)
        # now the sfac and unit tables are written to the resfile
        if not explicit_scat:
            self._reslist[sfacline] = 'SFAC  {}\n'.format('  '.join(sfac))
            self._reslist[unitline] = 'UNIT  {}\n'.format('  '.join(unit))  # builds the UNIT line
        return sfac + explicit_scat


#############################################################################


class Elem_2_Sfac():
    def __init__(self, sfac_table):
        """
        SFAC2Element and vice versa conversion
        :param sfac_table: SFAC table of the res file.
        :type sfac_table: list
        """
        self._sfac_table = sfac_table

    def elem_2_sfac(self, atom_type):
        """
        returns an sfac-number for the element given in "atom_type"
        :param atom_type: string 'C'
        :type atom_type: string

        >>> el = Elem_2_Sfac(['Ca', 'H', 'C', 'BR'])
        >>> el.elem_2_sfac('Br')
        4
        >>> el.elem_2_sfac('c')
        3
        >>> el.elem_2_sfac('Ar')
        """
        for num, element in enumerate(self._sfac_table, 1):
            if atom_type.upper() == element.upper():
                return num  # return sfac number
                break

    def sfac_2_elem(self, sfacnum):
        """
        returns an element and needs an sfac-number
        :param sfacnum: string like '2'
        >>> el = Elem_2_Sfac(['Ca', 'H', 'C', 'BR'])
        >>> el.sfac_2_elem(1)
        'CA'
        >>> el.sfac_2_elem(2)
        'H'
        >>> el.sfac_2_elem(3)
        'C'
        >>> el.sfac_2_elem(4)
        'BR'
        >>> el.sfac_2_elem(5)

        >>> el.sfac_2_elem(0)

        """
        try:
            sfacnum = int(sfacnum)
        except ValueError:
            return False
        for num, element in enumerate(self._sfac_table, 1):
            if sfacnum == num:
                return element.upper()  # return Element name
                break


#############################################################################


class NumberScheme():
    """
    This class needs a dbentry and the resfile contents to generate a unique
    numbering scheme for the dbentry.
    It essentially adds a letter from the alphabet to the atom name and iterates
    over the names until the names are unique.
    """

    def __init__(self, reslist, dbatome, dsrp):
        """
        :param reslist: resfile as list
        :param dbatome: list of atoms as list
        :param resi: residue True or false
        """
        self.dsrp = dsrp
        self._reslist = reslist
        self.__rlist = []
        for i in get_atoms(self._reslist):
            if not re.match('H[0-9]+', i[0]):
                if not re.match('Q[0-9]+', i[0]):
                    self.__rlist.append(i[0])
        self.__dbatome = dbatome
        # Alphabet for the naming scheme:
        self.__alphabet = []
        self.__letters = string.ascii_uppercase
        self.__alphabet = [i for i in string.ascii_uppercase]
        # unsorted atomlist:
        self.__atyp = get_atomtypes(self.__dbatome)
        # sorted atomlist:
        self.__atomtype = sorted(self.__atyp[:])

    def _get_number_of_at_in_type(self, atomtype, check):
        """
        returns the number of atoms which have the same type
        """
        num = 0  # @UnusedVariable
        num = atomtype.count(check)
        return num

    def _generate_numbers(self, atype, num, suffix=''):
        """
        generates numberscheme
        """
        atlist = []
        if num == 1:  # in case of only one atom of this type
            atlist.append("%s%s%s" % (atype, num, suffix))
            return atlist
        for i in range(1, num + 1):
            atlist.append("%s%s%s" % (atype, str(i), suffix))
        return atlist

    @staticmethod
    def _contains(atomlist, rlist):
        """
        checks if one of the atoms in the list exist in resfile
        """
        x = [e for e in atomlist if e in '\n'.join(rlist)]
        if x:
            return False
        else:
            return True

    def get_fragment_number_scheme(self, extranames=None):
        """
        returns a list of atoms of length len(self.__dbatome) whith a
        naming scheme which fits into the resfile. extraname is a list of atoms
        which should be also excluded from the new names.
        :param extranames: list of excluding atoms e.g. ['O1A']
        :type extranames: list
        """
        if not extranames:
            extranames = []
        self.__rlist.extend(extranames)
        if self.dsrp.resiflag:
            print('RESI instruction is enabled. Leaving atom numbers as they are.')
            return [i[0].upper() for i in self.__dbatome]
        num = 0
        atom = []
        newatom = []
        atomlist = []
        for x, i in enumerate(self.__atomtype):
            alph = self.__alphabet[:]
            if atom == i:
                continue
            atom = i
            # number of atoms with given atomtype:
            num = self._get_number_of_at_in_type(self.__atomtype, atom)  # unittest: atom instead of atomtype
            # list of atoms with given atomtype
            atomlist = self._generate_numbers(atom, num)
            # important if we want to have all atoms with the same suffix:
            self.__rlist.extend(atomlist)
            # check if rlist contains an atom name of atomlist.
            # If so, add a suffix from aplhabet to it.
            while not self._contains(atomlist, self.__rlist):
                suffix = alph[0]
                del alph[0]
                atomlist = self._generate_numbers(atom, num, suffix)
            newatom.extend(atomlist)
        # retain the original order of the atomlist:
        # index of the sorted atoms:
        aindex = sorted(list(range(len(self.__atyp))), key=lambda k: self.__atyp[k])
        orglist = [''] * len(newatom)
        for x, i in enumerate(aindex):
            orglist[i] = newatom[x]  # every ith aindex element is replaced by newatom[x] to retain original order
        return orglist


if __name__ == '__main__':
    import doctest

    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))
    else:
        print('{} of {} tests failed'.format(failed, attempted))



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#
import dataclasses
import os
import re
import sys
import tarfile
from copy import deepcopy
from pathlib import Path
from typing import List, Dict, Union

from atomhandling import get_atomtypes
from atoms import Element
from constants import atomregex, SHX_CARDS, RESTRAINT_CARDS, sep_line
from misc import atomic_distance, nalimov_test, std_dev, median, pairwise, \
    unwrap_head_lines, dice_coefficient2, frac_to_cart, read_cif

not_existing_error = '*** Fragment "{}" not found in database ***'


def invert_atomic_coordinates(atoms):
    """
    Inverts SHELXL atom coordinates.

    :param atoms: list of atom list

    >>> atoms = [['C1', 1, 0.44, 0.21, -1.23 ], ['C2', 1, -1.44, -0.21, 21.23]]
    >>> invert_atomic_coordinates(atoms)
    [['C1', 1, -0.44, -0.21, 1.23], ['C2', 1, 1.44, 0.21, -21.23]]
    >>> invert_atomic_coordinates([['C2', 1, 'd', 1, 2]])
    Unable to invert fragment coordinates.
    []
    """
    catoms = deepcopy(atoms)
    for line in catoms:
        try:
            inv_coord = [-x for x in line[-3:]]
        except Exception:
            print('Unable to invert fragment coordinates.')
            return []
        line[-3:] = inv_coord
    return catoms


def search_fragment_name(search_string, gdb, numresults=6):
    """
    searches the Name: comments in the database for a given name

    :param numresults: number of results to return after search
    :type search_string: str
    :param gdb: database object
    :type gdb: global_DB
    """
    names_list = []
    for fragment in gdb.databases:
        fragname = gdb[fragment]['name']
        line_number = gdb[fragment]['startline']
        dbname = gdb[fragment]['dbname']
        names_list.append([fragment, fragname, line_number, dbname])
    search_results = []
    for fragment in names_list:
        key = make_sortkey(fragment[1], searchkey=True)
        # key[0] and key[1] combined to have also the sum formula:
        fullstring = f"{key[0]}{key[1]}"
        if fragment[1].lower().startswith(search_string):
            # give names starting with a certain string high priority
            coefficient = 3.0
        elif search_string in fullstring or search_string in fragment[0]:
            coefficient = 2.1
        else:
            coefficient = dice_coefficient2(search_string, fullstring)
        fragment.append([coefficient, key[1]])
        if coefficient > 0.1:
            search_results.append(fragment)
    # select the best n results, sort for first and second search key
    selected_results = sorted(search_results, key=lambda coeff: [coeff[-1][0], coeff[-1][1]], reverse=True)[
                       :numresults]
    return selected_results


def print_search_results(results):
    """
    prints the results of a database search to screen and exit.
    results are
    """
    print(' Fragment          | Full name, Comments                      | Line number')
    print(sep_line)
    for line in results:
        print(f' {line[0]:15s}   | {line[1]:40s} | {line[2]}')
    sys.exit()


def make_sortkey(full_name, searchkey=False):
    """
    Algorythm inspired by W. Sage J. Chem. Inf: Comput. Sci. 1983, 23, 186-197
    """
    keylist = []
    full_name = ''.join(e for e in full_name if e not in ('{}()[],'))
    full_name = full_name.split(' ')
    rest = " " + ' '.join(full_name[1:])
    full_name = full_name[0].lower()
    numbers = ''.join(e for e in full_name if e in r'0123456789')
    if full_name.startswith('tert-'):
        full_name = full_name[4:]
    if full_name.startswith('sec-'):
        full_name = full_name[3:]
    if full_name.startswith('iso-'):
        full_name = full_name[3:]
    if full_name.startswith('bis-'):
        full_name = full_name[3:]
    if full_name.startswith('tris-'):
        full_name = full_name[3:]
    if full_name.startswith('mono-'):
        full_name = full_name[4:]
    if full_name.startswith('t-'):
        full_name = full_name[1:]
    if full_name.startswith('p-'):
        full_name = full_name[2:]
    if full_name.startswith('o-'):
        full_name = full_name[1:]
    if full_name.startswith('n-'):
        full_name = full_name[1:]
    if full_name.startswith('nn-'):
        full_name = full_name[1:]
    if full_name.startswith('nnn-'):
        full_name = full_name[1:]
    if full_name.startswith('m-'):
        full_name = full_name[1:]
    if full_name.startswith('i-'):
        full_name = full_name[1:]
    full_name = ''.join(e for e in full_name if e not in ('+-_.\'1234567890, '))
    if searchkey:
        keylist = [full_name + rest, numbers]  # enables search for sum formulae
    else:
        keylist = [full_name, numbers]
    return keylist


# Names of the database files:
# dsr_db.txt is the db from the distribution. This file should not be edited.
# dsr_user_db.txt if this file exists, all its content is also read in.

def read_file_data(filepath: str) -> List[str]:
    """
    reads the database files and returns them as list.
    """
    dblist = []
    try:
        with open(filepath, 'r') as f:
            dblist = f.read().splitlines()
    except (IOError, TypeError) as e:
        print(e)
    return dblist


class ParseDB():
    """
    reads in the system and user db files and makes
    a dictionary of them.
    maindb = read_file_data(path_to_maindb)
    """

    def __init__(self, maindb_path: str = '', userdb_path: str = ''):
        """
        self._databases: dictionary with the individial fragments
        """
        self.databases = {}
        if maindb_path:
            self.parse(maindb_path, 'dsr_db')
        if userdb_path:
            self.parse(userdb_path, 'dsr_user_db')

    def parse(self, dbpath: str = '', dbname: str = 'dsr_db') -> Dict[str, Dict[str, str]]:
        """
        This method returns all fragment name tags in the database

        >>> dbpath = os.path.abspath('./src/dsr_shelx/dsr_db.txt')
        >>> db = ParseDB(dbpath)
        >>> db.parse(dbpath, 'dsr_db')['water']['name'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        'Water, H2O'
        >>> db.parse(dbpath, 'dsr_db')['water']['comments']
        []
        >>> db.parse(dbpath, 'dsr_db')['water']['atoms'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        [['O1', 4, 0.0, 0.0, 0.0],
        ['H1', 2, 0.9584, 0.0, 0.0],
        ['H2', 2, -0.2392, 0.9281, 0.0]]
        >>> db.parse(dbpath, 'dsr_db')['water']['cell'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        [1.0, 1.0, 1.0, 90.0, 90.0, 90.0]
        >>> db.parse(dbpath, 'dsr_db')['water']['source'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        'pbe1pbe/6-311++G(3df,3pd), Ilia A. Guzei'
        >>> db.parse(dbpath, 'dsr_db')['water']['resi'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        'H2O'
        >>> db.parse(dbpath, 'dsr_db')['water']['restraints'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ['DFIX 0.9584 0.001 O1 H1 O1 H2', 'DFIX 1.5150 0.001 H1 H2']
        >>> db.parse(dbpath, 'dsr_db')['water']['dbname'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        'dsr_db'
        >>> db.parse(dbpath, 'dsr_db')['water']['hfix'] # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        []
        """
        frag_tag = ''
        db: Dict[str, Dict[str, str]] = dict()
        start_regex = re.compile(r'<[^/].*>', re.IGNORECASE)  # regular expression for db tag.
        starttag = False
        fraglines = []
        end_regex = None
        startnum = 0
        for num, line in enumerate(read_file_data(dbpath)):
            if line.startswith('#'):
                continue
            if not line:
                continue
            # matching end tag
            if end_regex and end_regex.match(line):
                starttag = False
                db[frag_tag].update(dict(endline=num + 1, startline=startnum + 1))
                db = self.parse_fraglines(frag_tag, fraglines, db)
                fraglines = []
            # start tag was found, appending lines to fragment list
            if starttag:
                fraglines.append(line)
            # matching start tag and compiling end regex
            if start_regex.match(line):
                if starttag:
                    print(f'*** Error in database "{dbname}.txt" in line {num + 1}. End tag is missing ***')
                # lower case is essential here:
                frag_tag = line.strip('<> \n\r').lower()
                if frag_tag in db:
                    print('\n*** Duplicate database entry "{}" found! Please remove/rename '
                          'second entry\nand/or check all end tags in the database dsr_usr_db.txt '
                          'or dsr_db.txt. ***'.format(frag_tag))
                    sys.exit()
                starttag = True
                startnum = num
                db[frag_tag] = {'dbname': dbname}
                end_regex = re.compile(re.escape(r'</{}>'.format(frag_tag)), re.IGNORECASE)
        self.databases.update(db)
        return db

    @staticmethod
    def parse_fraglines(fragname_tag: str, fraglines: list, db: dict) -> dict:
        """
        Fills the database dictionary with fragment data.
        """
        fragname_tag = fragname_tag.lower()
        headlist = []
        comments = []
        hfix = []
        residue = ''
        atoms = []
        cell = []
        source = ''
        name = ''
        nameregex = re.compile(r'REM\s+NAME:', re.IGNORECASE)
        srcregex = re.compile(r'REM\s+(SRC:|SOURCE:)', re.IGNORECASE)
        # devide atoms and the rest:
        for num, aline in enumerate(fraglines):
            if atomregex.match(aline):  # search atoms
                atline = aline.split()[:5]  # convert to list and use only first 5 columns
                if atline[0] not in SHX_CARDS:  # exclude all non-atom cards
                    coords = []
                    try:
                        coords = [float(x) for x in atline[2:]]
                        atline[1] = int(atline[1])
                    except ValueError:
                        print("*** Invalid atomic coordinates in line {} of {}.txt (Fragment: {}) ***"
                              .format(db[fragname_tag]['startline'] + num + 1, db[fragname_tag]['dbname'],
                                      fragname_tag))
                        sys.exit()
                    atline[2:] = coords
                    atoms.append(atline)
                    fraglines[num] = ''
        # bring all wrapped lines with = at end to a single line:
        fraglines = unwrap_head_lines(fraglines)
        for num, line in enumerate(fraglines):
            # collect the comments:
            if line.upper().startswith('REM'):
                if line.upper().startswith('REM HFIX'):
                    hfix.append(line)
                    continue
                # Where do the fragment come from?
                if srcregex.match(line):
                    source = ' '.join(line.split()[2:])
                    fraglines[num] = ''
                    continue
                # Get the full name:
                if nameregex.match(line):
                    name = ' '.join(line.split()[2:])
                    fraglines[num] = ''
                    continue
                comments.append(line)
                continue
            command = line[:4].upper().strip()
            if command and command in SHX_CARDS:
                if command == 'RESI':
                    resiline = line.split()
                    for n, i in enumerate(resiline):
                        if not i[0].isalpha():
                            del resiline[n]
                    try:
                        residue = resiline[1].upper()
                        continue
                    except IndexError:
                        print('*** Invalid residue definition in database entry {}. ***'.format(fragname_tag))
                        sys.exit()
                # collect the unit cell of the fragment:
                if command == 'FRAG':
                    try:
                        cell = [float(x) for x in line.split()[2:8]]
                    except (ValueError, IndexError):
                        print('*** Invalid unit cell found in database entry {}. ***'.format(fragname_tag))
                        sys.exit()
                    continue
                # these must be restraints:
                if line[:4] in RESTRAINT_CARDS:
                    headlist.append(line)
            elif command:
                print('*** Bad line {} in database entry "{}" found! ({}.txt) ***'
                      .format(num + db[fragname_tag]['startline'] + 1, fragname_tag, db[fragname_tag]['dbname']))
                print(line)
        if not cell:
            print('*** Error. No cell parameters or malformed cell found in the database entry ' \
                  'of "{}" ***'.format(fragname_tag))
            sys.exit()
        if not name:
            # This happens if no name is given or "REM Name:" has errors (The name is explicitely not mandatory!).
            name = fragname_tag
        if not db:
            print('*** No database found! ***\n')
            sys.exit()
        if not atoms:
            # Can not print this out, because shelXle GUI will fail if text is printed.
            # print('*** No atoms found in database entry {} line {} of {}.txt***'.format(fragname_tag,
            #                                    db[fragname_tag]['startline'] + 1, db[fragname_tag]['dbname']))
            del db[fragname_tag]
            return db
        db[fragname_tag].update({
            'restraints': headlist,  # header with just the restraints
            'resi'      : residue,  # the residue class
            'cell'      : cell,  # FRAG ...
            'atoms'     : atoms,  # the atoms as lists of list
            'comments'  : comments,  # the comment line
            'source'    : source,
            'name'      : name,
            'hfix'      : hfix,
        })
        return db

    def __getitem__(self, fragment):
        # type: (str) -> dict
        try:
            return self.databases[fragment]
        except KeyError:
            print("*** Fragment {} was not found in the Database! ***".format(fragment))
            sys.exit()

    def __iter__(self):
        """
        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> [x for x in db][:3]
        ['12-dichlorobenz', '12-difluorobenz', '12c4']
        """
        # return iter(self.databases.keys()) #python2
        return iter(list(sorted(self.databases.keys())))

    def list_fragments(self) -> List[List[str]]:
        """
        list all fragments in the db as list of lists
        [['tbu-c', 1723, 'dsr_db', 'Tert-butyl-C'], ...]
        """
        fraglist = []
        for frag in self.databases:
            name = self.get_fragment_name(frag)
            key = make_sortkey(name)
            line = [frag,
                    self.get_startline(frag),
                    self.get_db_name(frag),
                    name,
                    key[0],
                    key[1]]
            fraglist.append(line)
        fraglist.sort(key=lambda x: (x[4], x[5]))
        return fraglist

    def get_atomic_numbers(self, fragment: str = '') -> List[str]:
        """
        Returns the atomic numbers of the atoms in a fragment in
        same order as in the database as list.
        """
        atnumbers = []
        types = get_atomtypes(self.get_atoms(fragment))
        el = Element()
        for i in types:
            atnumbers.append(el.get_atomic_number(i))
        return atnumbers

    def get_atomnames(self, fragment='', uppercase=False):
        # type: (str, bool) -> list
        """
        Returns a list of only the atom names of a fragment.
        """
        atoms = self.get_atoms(fragment)
        names = []
        for i in atoms:
            if uppercase:
                names.append(i[0].upper())
            else:
                names.append(i[0])
        return names

    def get_head_for_gui(self, fragment):
        """
        returns header information of the specific fragment:
        tag, Name/comment, source, cell, residue, dbtype, restr, atoms
        >>> dbpath = os.path.abspath('./src/dsr_shelx/dsr_db.txt')
        >>> db = ParseDB(dbpath)
        >>> db.get_head_for_gui('benZene')
        ... # doctest: +NORMALIZE_WHITESPACE
        <tag>
         benzene
        </tag>
        <comment>
         Benzene, Benzol, Phenyl, C6H6
        </comment>
        <source>
         UGEDEQ
        </source>
        <cell>
         1;;1;;1;;90;;90;;90
        </cell>
        <residue>
         BENZ
        </residue>
        <dbtype>
         dsr_db
        </dbtype>
        <restr>
         SADI 0.02 C1 C2 C2 C3 C3 C4 C4 C5 C5 C6 C6 C1;;SADI 0.04 C1 C5 C1 C3 C5 C3 C4 C2 C4 C6 C6 C2;;FLAT C1 > C6;;SIMU C1 > C6;;RIGU C1 > C6
        </restr>

        """
        fragment = fragment.lower()
        print("<tag>\n", fragment, "\n</tag>")
        print("<comment>\n", self.get_fragment_name(fragment), "\n</comment>")
        print("<source>\n", self.get_src(fragment), "\n</source>")
        # it returns cartesian coordinates, so we need this cell here:
        print("<cell>\n", '1;;1;;1;;90;;90;;90', "\n</cell>")
        print("<residue>\n", self.get_resi(fragment), "\n</residue>")
        print("<dbtype>\n", self.get_db_name(fragment), "\n</dbtype>")
        print("<restr>\n", ';;'.join(self.get_restraints(fragment)), '\n</restr>')
        self.check_consistency(fragment)
        self.check_db_restraints_consistency(fragment)
        if not self.check_sadi_consistence(fragment):
            sys.exit()
        self.check_db_atom_consistency(fragment)

    def check_consistency(self, fragment):
        # type: (str) -> bool
        """
        check if the fragline makes sense and if the fragment_dict is complete
        """
        fragment = fragment.lower()
        if fragment in ['cf3', 'cf6', 'cf9']:
            return True
        try:
            dbentry = self.databases[fragment]
        except KeyError:
            print("*** Fragment {} not found in the database. ***".format(fragment))
            sys.exit()
        if not dbentry['cell']:
            print('*** Error. No cell parameters or malformed cell found in the database entry ' \
                  'of "{}" ***'.format(fragment))
            print('*** Please add these parameters! ***')
            sys.exit()
        if not dbentry['restraints']:
            print('*** Restraints in the header of database entry "{}" ({}) missing! Check your Database! ***' \
                  .format(fragment, dbentry['name']))
            return False
        if not dbentry['resi']:
            print('*** Residue in the header of database entry "{}" ({}) missing! Check your Database! ***' \
                  .format(fragment, dbentry['name']))
        if not dbentry['atoms']:
            print('*** No atoms found in database entry {} line {} of {}.txt***'.format(fragment,
                                                                                        dbentry['startline'] + 1,
                                                                                        dbentry['dbname']))
            print('*** Have you really followed the syntax? ***')
            sys.exit()
        if not dbentry['endline']:
            print('*** Could not find end of dbentry for fragment "{}" in line {} of "{}.txt". '
                  'Check your database. ***'.format(fragment, dbentry['startline'] + 1, dbentry['dbname']))
            sys.exit()
        return True

    def get_sum_formula(self, fragment):
        """
        returns the sum formula of a fragment
        as string like 'SI4 C9 O1'
        :param fragment: fragment name
        :type fragment: string
        """
        fragment = fragment.lower()
        try:
            types = get_atomtypes(self.databases[fragment]['atoms'])
        except Exception:
            return None
        formula = ''
        for el in set(types):
            num = types.count(el)
            formula += '{}{} '.format(el, num)
        return formula

    def check_db_atom_consistency(self, fragment):
        # type: (str) -> bool
        """
        This method is for atoms only (without db header)!
        check the db for duplicates:
        """
        dbatoms = self.get_atomnames(fragment, uppercase=True)
        # check for duplicates:
        while dbatoms:
            at = dbatoms.pop()
            if at in dbatoms:
                print('*** Duplicate atom {0} in database entry "{1}" ({2}) '
                      'found! Check your database... ***'
                      .format(at, fragment, self.databases[fragment]['name']))
                sys.exit()
        return True

    def get_hfixes(self, fragment, resi_class):
        """
        Returns pre-defined hfix instructions.
        :param fragment: fragment name
        :return: A string with the REM HFIX ... instruction
        >>> from src.dsr_shelx.misc import wrap_headlines
        >>> hf = ['REM HFIX 123 C1 C2 C3 C4 C5']
        >>> wrap_headlines(hf, 18)
        ['REM HFIX 123 C1 C2 =\\n   C3 C4 C5\\n']
        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> db.get_hfixes('Adamantane', 'ADAM')
        ['REM HFIX_ADAM 23 C2 C3 C4 C6 C8 C10', 'REM HFIX_ADAM 13 C5 C7 C9']
        >>> db.get_hfixes('Adamantane', '')
        ['REM HFIX 23 C2 C3 C4 C6 C8 C10', 'REM HFIX 13 C5 C7 C9']
        >>> db.get_restraints('Adamantane')
        ['SADI C1 C2 C1 C3 C1 C4 C5 C6 C6 C7 C7 C8 C8 C9 C9 C10 C10 C5 C2 C9 C3 C7 C4 C5', 'SADI 0.04 C2 C3 C3 C4 C4 C2 C5 C9 C9 C7 C7 C5 C6 C10 C6 C8 C8 C10', 'SADI 0.05 C5 C8 C7 C10 C9 C6', 'SIMU C1 > C10', 'RIGU C1 > C10']
        >>> db.get_hfixes('oc(cf3)3', '')
        []
        """
        fragment = fragment.lower()
        try:
            hfix = [' '.join(x.upper().split()) for x in self.databases[fragment]['hfix']]
        except KeyError:
            print(not_existing_error.format(fragment))
            sys.exit()
        if hfix and resi_class:
            resihfix = []
            for line in hfix:
                # Adding _resiclass to HFIX instructions:
                # REM HFIX C1 C2  ->  REM HFIX_class C1 C2 
                line = '{}_{} {}'.format(line[:8], resi_class, line[9:])
                resihfix.append(line)
            return resihfix
        return hfix

    def check_db_restraints_consistency(self, fragment):
        """
        - Checks if the Atomnames in the restraints of the dbhead are also in
          the list of the atoms of the respective dbentry.
        - Checks wether restraints cards are vaid.
        """
        fragment = fragment.lower()
        atoms = self.get_atoms(fragment)
        status = True
        atoms = [i[0].upper().strip() for i in atoms]
        restraint_atoms_list = set([])
        for line in self.get_restraints(fragment):
            for i in line.split():
                # Todo: resolve ranges:
                if i in ('>', '<'):
                    continue
                try:
                    # Test if first parameter is a distance or a standard deviation:
                    if i in ['DFIX', 'DANG']:
                        # print('####')
                        if len(line.split()) > 1:  # there is more than just DFIX
                            # Test if this is correct: DFIX d s[0.02] atom pairs (d should be greater s)
                            if float(line.split()[1]) < float(line.split()[2]):
                                print('\n*** Sigma > d in DFIX/DANG of "{}: {}" '
                                      'in database {}. ***'
                                      .format(fragment, self.get_fragment_name(fragment), self.get_db_name(fragment)))
                        else:
                            print('\n*** Incomplete DFIX/DANG in restraints of "{}: {}" in database {}. ***'
                                  .format(fragment, self.get_fragment_name(fragment), self.get_db_name(fragment)))
                    # if i in SHX_CARDS:
                    #    continue
                    else:
                        # just test if there is not a number:
                        float(i)
                except ValueError:
                    if i in SHX_CARDS:
                        continue
                    restraint_atoms_list.add(i)
        for atom in restraint_atoms_list:
            if atom.upper() not in atoms:
                status = False
                print('\n*** Unknown atom "{}" in restraints of "{}: {}" in database {}. ***'
                      .format(atom, fragment, self.get_fragment_name(fragment), self.get_db_name(fragment)))
        if not status:
            print('*** Check database entry. ***\n')
            sys.exit()
        return status

    def check_sadi_consistence(self, fragment):
        """
        check if same distance restraints make sense. Each length of an atom
        pair is tested agains the standard deviation of all distances.
        For a large standard deviation, the list is tested for outliers.
        :param fragment: frag name
        """
        atoms = self.get_atoms(fragment)
        restr = self.get_restraints(fragment)
        restraints = deepcopy(restr)
        atnames = self.get_atomnames(fragment, uppercase=True)
        good = True
        for num, line in enumerate(restraints):
            prefixes = []
            dev = 0.02
            line = line.split()
            if not line:
                continue
            if line[0].upper() == 'SADI':
                prefixes.append(line[0])
                del line[0]
                try:
                    if not str(line[0][0]).isalpha():
                        prefixes.append(line[0])
                        dev = line[0]
                        del line[0]  # delete standard deviation
                except IndexError:
                    return False
                if len(line) % 2 == 1:  # test for uneven atoms count
                    print('*** Inconsistent SADI restraint line {} of "{}". '
                          'Not all atoms form a pair ***'.format(num, fragment))
                pairs = pairwise(line)
                distances = []
                pairlist = []
                if len(pairs) <= 1:
                    return True
                for i in pairs:
                    if i in pairlist or tuple(reversed(i)) in pairlist:
                        print('*** Duplicate atom pair "{}" in SADI restraint line {} of "{}". ***'.format(" ".join(i),
                                                                                                           num,
                                                                                                           fragment))
                    pairlist.append(i)
                    try:
                        atom1 = [float(x) for x in atoms[atnames.index(i[0])][2:5]]
                        atom2 = [float(x) for x in atoms[atnames.index(i[1])][2:5]]
                    except ValueError:
                        return False
                    dist = atomic_distance(atom1, atom2, self.get_cell(fragment))
                    distances.append(dist)
                if len(pairlist) == 2:
                    # Find restraints with one pair, where 1,2 and 1,3 distances are mixed:
                    pairdev = 1.0 - (min(distances) / max(distances))
                    if pairdev > (2.0 * float(dev)):
                        print('*** Suspicious deviation of {:.3f} A for "{}" in {} ***'.format(pairdev, restraints[num],
                                                                                               fragment))
                        return False
                    return True  # because esd is too sensitive for two distances.
                stdev = std_dev(distances)  # Error distribution of
                # only do outlier test if standard deviation is suspiciously large:
                if stdev > 0.065:
                    outliers = nalimov_test(distances)
                    if outliers:
                        print("\nFragment {}:".format(fragment))
                        for x in outliers:
                            pair = ' '.join(pairlist[x])
                            print('*** Suspicious deviation of atom pair "{}" ({:4.3f} A, median: {:4.3f}) after '
                                  'line {} in {}.txt ***'.format(pair, distances[x], median(distances),
                                                                 self.get_startline(fragment) + num + 1,
                                                                 self.get_db_name(fragment))
                                  )
                            print('*** {} ... ***'.format(restr[num][:60]))
                            good = False
                if (stdev > (2.5 * float(dev))) and good:
                    print("\nFragment {}:".format(fragment))
                    print(
                        '*** Suspicious restraints in SADI line {} with high standard deviation {:4.3f} '
                        '(median length: {:4.3f} A) ***'.format(num + 1, stdev, median(distances)))
                    print('*** ' + restraints[num] + ' ***\n')
                    good = False
        if good:
            return True
        else:
            return False

    def search_for_error_response(self, fragment):
        """
        searches for a fragment name in the db as response to an invalid fragment name.
        :param fragment: the fragment
        :type fragment: string
        """
        result = search_fragment_name(fragment.lower(), self.databases)
        print('Do you mean one of these?:\n')
        print_search_results(result)

    def get_atoms(self, fragment, invert=False, cartesian=False):
        # type: (str, bool, bool) -> list
        """
        returns the atoms from the dbentry:
        [['O1', '1', '0.01453', '-1.6659', '-0.10966'],
        ['C1', '1', '0.00146', '-0.26814', '-0.06351'], ... ]
        :param fragment: fragment name
        :type fragment: string

        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> db.get_atoms('water', cartesian=False) # doctest: +NORMALIZE_WHITESPACE
        [['O1', 4, 0.0, 0.0, 0.0], ['H1', 2, 0.9584, 0.0, 0.0], ['H2', 2, -0.2392, 0.9281, 0.0]]
        >>> db.get_atoms('isoprop', cartesian=True) # doctest: +NORMALIZE_WHITESPACE
        [['O1', 1, 5.611626689944858, 4.650686219462244, 9.549527472636036],
        ['C1', 1, 5.969103603542006, 3.2932494073414893, 9.793246423730249],
        ['C2', 1, 4.814135175316713, 2.3627923165385942, 9.795180701119886],
        ['C3', 1, 7.0650724256725805, 2.9180586416449548, 8.82997628369121]]
        """
        fragment = fragment.lower()
        try:
            atoms = deepcopy(self.databases[fragment]['atoms'])
        except KeyError:
            print('*** Could not find {} in database ***'.format(fragment))
            sys.exit()
        if cartesian:
            cell = self.get_cell(fragment)
            for num, c in enumerate(self.get_coordinates(fragment)):
                atoms[num][2:5] = frac_to_cart(c, cell)
        if invert:
            atoms = invert_atomic_coordinates(atoms)
        return atoms

    def get_cell(self, fragment):
        # type: (str) -> list
        """
        returns the line with FRAG 17 cell from the dbentry
        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> db.get_cell('water')
        [1.0, 1.0, 1.0, 90.0, 90.0, 90.0]
        >>> db.get_cell('foobar')
        Traceback (most recent call last):
        ...
        SystemExit
        """
        try:
            cell = [float(x) for x in self.databases[fragment.lower()]['cell']]
        except KeyError:
            print(not_existing_error.format(fragment))
            sys.exit()
        return cell

    def get_startline(self, fragment):
        # type: (str) -> int
        """
        returns the line number from the dbentry
        """
        try:
            return int(self.databases[fragment.lower()]['startline'])
        except KeyError:
            print(not_existing_error.format(fragment))
            return 0
        except ValueError:
            print('Could not find startline of Fragment {}'.format(fragment))
            return 0

    def get_coordinates(self, fragment, cartesian=False, invert=False):
        # type: (str, bool, bool) -> list
        """
        Returns the coordinates of the fragment. Optionally is direct conversion to cartesian coordinates.
        [['0.01453', '-1.6659', '-0.10966'], [...] ]
        :param fragment:
        :param cartesian:
        :return:
        """
        try:
            atoms = self.get_atoms(fragment, invert=invert, cartesian=cartesian)
            coords = [i[2:5] for i in atoms]
        except KeyError:
            print(not_existing_error.format(fragment))
            sys.exit()
        return coords

    def get_restraints(self, fragment: str) -> List[str]:
        """
        returns the header of the dbentry of fragment.
        This header does not include comments, only the restraints.
        """
        fragment = fragment.lower()
        try:
            restr = [x.upper() for x in self.databases[fragment]['restraints']]
        except KeyError:
            print(not_existing_error.format(fragment))
            sys.exit()
        return restr

    def get_resi(self, fragment):
        """
        returns the residue name of the dbentry of fragment
        can be either class or class + number.
        convention is only class.
        """
        try:
            return self.databases[fragment.lower()]['resi']
        except KeyError:
            print(not_existing_error.format(fragment))
            return ''

    def get_fragment_name(self, fragment):
        """
        returns the first comment line of the dbentry of a fragment
        if a line with "rem Name:" is present, this line is used as comment.
        :param fragment: actual fragment name
        :type fragment: string
        """
        try:
            name = self.databases[fragment.lower()]['name']
        except KeyError:
            print(not_existing_error.format(fragment))
            return ''
        return name

    def get_src(self, fragment):
        """
        returns the source line of the dbentry of a fragment
        if a line with "rem Src:" is present.
        :param fragment: actual fragment name
        :type fragment: string
        """
        try:
            src = self.databases[fragment.lower()]['source']
        except KeyError:
            print(not_existing_error.format(fragment))
            return ''
        return src

    def get_db_name(self, fragment):
        """
        returns the fragment database name of fragment x
        """
        return self.databases[fragment.lower()]['dbname']

    def get_fragment_tags(self):
        tags = [x[0] for x in self.list_fragments()]
        return tags


def get_first_last_atom(atoms):
    """
    :type atoms: list
    returns the first and the last atom from the imported atom list
    """
    try:
        first = atoms[0][0]
        last = atoms[-1][0]
    except (KeyError, IndexError):
        return False
    return first, last


@dataclasses.dataclass()
class GradeFiles():
    cif_file: str
    pdb_file: str
    dfix_file: str
    dfix_file_hydrogen: str


class ImportGRADE():
    def __init__(self, grade_tar_file: str, db, invert=False, maindb=None, userdb=None):
        """
        class to import fragments from GRADE of Global Phasing Ltd.
        :type maindb:   str, None
        :param userdb:  directory where the user database is located.
                        Default is the users home directory.
        :type userdb:   string
        """
        self.user_db_path = userdb
        self.main_db_path = maindb
        self.el = Element()
        self.invert = invert
        self._db = db
        self._db_dir = os.path.expanduser("~")
        self._db_tags = self._db.get_fragment_tags()
        self._db = self._db.databases
        self.gradefiles = self.get_gradefiles(Path(grade_tar_file))
        self._atoms = self.get_pdbatoms(self.gradefiles.pdb_file)
        self._firstlast = get_first_last_atom(self._atoms)
        self._restraints = self.get_restraints()
        self._resi_name = self.get_resi_from_pdbfile()
        if not isinstance(self._resi_name, str):
            self._resi_name = self._resi_name.decode()

    def get_gradefiles(self, grade_tar_file: Path) -> GradeFiles:
        """
        returns the .mol2 and .dfix file location
        """
        if grade_tar_file.suffix == '.tgz':
            try:
                gradefile = tarfile.open(grade_tar_file)  # , encoding="ascii")
            except IOError:
                print('No such file or directory: {}'.format(grade_tar_file))
                sys.exit()
        else:
            print(f'*** File {grade_tar_file} is not a valid file to import from ***')
            sys.exit()
        pdbfile = ''
        dfixfile = ''
        dfixfile_hydrogens = ''
        cif_file = ''
        for filename in gradefile.getnames():
            if filename.endswith('.pdb'):
                if re.match(r'.*obabel.*', filename):
                    continue
                pdbfile = gradefile.extractfile(filename)
            if filename.endswith('.dfix') and 'with_hydrogen' not in Path(filename).suffixes:
                dfixfile = gradefile.extractfile(filename)
            if filename.endswith('.dfix') and 'with_hydrogen' in Path(filename).suffixes:
                dfixfile_hydrogens = gradefile.extractfile(filename)
            if filename.endswith('.cif'):
                cif_file = gradefile.extractfile(filename)
        if not pdbfile:
            print(' .pdb file not found!')
            self.import_error(grade_tar_file)
        elif not dfixfile:
            print(' .dfix file not found!')
            self.import_error(grade_tar_file)
        output = GradeFiles(pdb_file=pdbfile.read().decode('latin1'),
                            dfix_file=dfixfile.read().decode('latin1'),
                            dfix_file_hydrogen=dfixfile_hydrogens,
                            cif_file=cif_file.read().decode('latin1'))
        return output

    @staticmethod
    def _read_tarfile_as_ascii(filehandle):
        tmp = []
        for line in filehandle:
            if not isinstance(line, str):
                line = line.decode('ascii')
            tmp.append(line)
        return tmp

    def get_name_from_pdbfile(self) -> str:
        """
        get the fragment name from the pdbfile.txt file

        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> mog = ImportGRADE('./tests/test-data/ALA.gradeserver_all.tgz', db)
        >>> mog.get_name_from_pdbfile()
        'Alanine'
        """
        full_name = ''
        full_name_regex = re.compile(r'^.*Compound full name.*')
        for line in self.gradefiles.pdb_file.splitlines(keepends=False):
            if not isinstance(line, str):
                line = line.decode('ascii')
            if not line:
                continue
            if full_name_regex.match(line):
                line = line.replace('_', '')
                line = line.replace('-', '')
                line = line.replace('#', '')
                try:
                    line = line.split()
                    _ = line[4]
                except IndexError:
                    full_name = ''
                else:
                    full_name = line[5]
        return full_name

    def get_name_from_cif_file(self) -> str:
        name = ''
        cif_str = self.gradefiles.cif_file.splitlines(keepends=False)
        cif = read_cif(cif_str)
        cif_value = cif.get('_chem_comp.name')
        if cif_value and isinstance(cif_value, (list, tuple)):
            name = cif_value[0]
        return name

    def get_resi_from_pdbfile(self):
        """
        get the fragment name from the pdbfile.txt file

        >>> db = ParseDB('./src/dsr_shelx/dsr_db.txt')
        >>> mog = ImportGRADE('./tests/test-data/ALA.gradeserver_all.tgz', db)
        >>> mog.get_resi_from_pdbfile()
        'ALA'
        """
        resi_name = None
        resi_regex = re.compile(r'^HETATM\s+1.*')
        for line in self.gradefiles.pdb_file.splitlines(keepends=False):
            if not isinstance(line, str):
                line = line.decode('ascii')
            if not line:
                continue
            if resi_regex.match(line):
                line = line.replace('_', '')
                line = line.replace('-', '')
                line = line.replace('#', '')
                resi_name = line.split()[3]
                break
        return resi_name

    def get_comments(self):
        """
        returns a detailed comment out of the .dfix file where the fragment was imported from

        #   REM Produced by Grade Web Server http://grade.globalphasing.org
        #   REM GEN: Generated by GRADE 1.2.5 (December 20 2013)
        #   REM GEN: from SMILES C1=CC=C2C(=C1)C=C(C3=CC=CC=C23)Br
        #   REM GEN: using MOGUL 1.6(RC5), CSD as535be with quantum mechanics RM1
        #   REM grade-cif2shelx output
        #   REM Version: 0.0.5 <Dec 20 2013>
        """
        matches = ['REM Produced by Grade', 'REM GEN:', 'REM grade-cif2shelx', 'REM Version:',
                   'REM Total charge', 'REM Name:']
        comments = []
        name = f'REM Name: {self.get_name_from_pdbfile() or self.get_name_from_cif_file()}'
        comments.append(name.split())
        for m in matches:
            for line in self.gradefiles.dfix_file.splitlines():
                if not isinstance(line, str):
                    line = line.decode()
                if line.startswith(m):
                    comments.append(line.split())
        return comments

    def get_restraints(self):
        """
        reads the restraints from a dfix-file
        """
        restraints = []
        for line in self.gradefiles.dfix_file.splitlines(keepends=False):
            if not isinstance(line, str):
                line = line.decode()
            line = line.split()
            if line:
                line[0] = line[0][:4].upper()
                if line[0] in RESTRAINT_CARDS:
                    restraints.append(line)
        if self._firstlast:
            atoms = ' > '.join(self._firstlast)
            atoms_s = 'SIMU ' + atoms
            atoms_s = atoms_s.split()
            atoms_r = 'RIGU ' + atoms
            atoms_r = atoms_r.split()
            restraints.append(atoms_s)
            restraints.append(atoms_r)
        return restraints

    def get_pdbatoms(self, pdb):
        """
        returns the atoms from a pdb file
        """
        atomlines = []
        for line in pdb.splitlines(keepends=False):
            if not isinstance(line, str):
                line = line.decode()
            line = line.split()
            if not line:
                continue
            if line[0] == 'HETATM':
                if line[-1] == 'H':
                    continue
                tmp = [line[2], line[11]]
                tmp.extend(line[6:9])
                atomlines.append(tmp)
        if not atomlines:
            self.import_error(pdb)
        if self.invert:
            atomlines = invert_atomic_coordinates(atomlines)
        return atomlines

    def bild_grade_db_entry(self):
        """
        builds a dbentry from the information supplied by GRADEs
        .mol2 and .dfix file
        """
        db_import_dict = {}
        num = 0
        name = self._resi_name[:4].upper()
        if not isinstance(name, str):
            name = name.decode()
        resi_name = name
        if not self._db_tags:
            print('*** Unable to import fragment. Database is empty. ***')
            sys.exit()
        # Check if anything is already in database:
        # tags are always lower case:
        while resi_name.lower() in self._db_tags:
            # Already there, so add a number to the tag:
            num = num + 1
            resi_name = resi_name[:3] + str(num)
        if num == 0:
            resi_name = resi_name[:3]
        else:
            resi_name = resi_name[:3] + str(num)
        fragline = 'FRAG 17 1  1  1  90  90  90'
        db_import_dict[resi_name] = {
            'restraints': self._restraints,
            'resi'      : resi_name,
            'cell'      : fragline.split(),
            'atoms'     : self._atoms,
            'line'      : None,
            'db'        : 'dsr_user_db',
            'comments'  : self.get_comments(),
            'name'      : resi_name
        }
        return db_import_dict

    @staticmethod
    def import_error(filename: Union[str, Path]):
        """
        warns for import errors
        """
        print(f'*** Unable to import GRADE file {filename} GRADE import to DSR relies on GRADE v1.100 and up. ***')
        sys.exit()

    def format_atomlist(self, atomlist: List[str]) -> List[str]:
        formatted_list = []
        for y in atomlist:
            atom_name = y[1]
            atom_name = ''.join([x for x in atom_name if x.isalpha()])
            formatted_string = f'{y[0]:<6} -{self.el.get_atomic_number(atom_name)}  {float(y[2]):>8.3f}' \
                               f'{float(y[3]):>8.3f}{float(y[4]):>8.3f}'
            formatted_list.append(formatted_string)
        return formatted_list

    def write_user_database(self):
        """
        writes content of existing dsr_user_db.txt and the imported GRADE entry to
        the dsr_user_db.txt
        """
        fragments = list(self._db)
        imported_entry = self.bild_grade_db_entry()
        grade_db_names = list(imported_entry)
        user_db_names = []
        for i in fragments:
            if self._db[i]['dbname'] == 'dsr_user_db':
                user_db_names.append(self._db[i]['name'])
        # try to write the new dbentry:
        try:
            with open(self.user_db_path, 'w') as f:
                for name in grade_db_names:
                    print(f'Importing {self._resi_name} ({name}) to user database...')
                    dbentry = self.compile_dbentry(imported_entry, name)
                    f.write(dbentry)
        except IOError as e:
            print(e)
            sys.exit()
        # try to write existing dbentries:
        try:
            with open(self.user_db_path, 'a+') as fu:
                for tag in fragments:
                    if self._db[tag]['dbname'] == 'dsr_user_db':
                        atomlist = self._db[tag]['atoms']
                        head = '\n'.join([''.join(x) for x in self._db[tag]['restraints']])
                        atoms = '\n'.join([f'{y[0]:<6}{y[1]:<2}{float(y[2]):>8.3f}'
                                           f'{float(y[3]):>8.3f}{float(y[4]):>8.3f}' for y in atomlist])
                        resi_name = self._db[tag]['resi']
                        comments = '\n'.join(self._db[tag]['comments'])
                        name = self._db[tag]['name']
                        comments = "REM Name: {}\n{}".format(name, comments)
                        fragline = 'FRAG 17 {} {} {} {} {} {}'.format(*self._db[tag]['cell'])
                        dbentry = (f'\n<{tag}> \n{comments} \nRESI {resi_name} \n'
                                   f'{head} \n{fragline} \n{atoms} \n</{tag}>\n')
                        fu.write(dbentry)
        except IOError as e:
            print(e)
            sys.exit()
        print('User database successfully updated.')

    def compile_dbentry(self, imported_entry: dict, name: str) -> str:
        atomlist = imported_entry[name]['atoms']
        comments = imported_entry[name]['comments']
        comments = '\n'.join([' '.join(i) for i in comments if i])
        head = '\n'.join([' '.join(x) for x in imported_entry[name]['restraints']])
        atoms = '\n'.join(self.format_atomlist(atomlist))
        resi_name = str(name)
        cell = '  '.join(imported_entry[name]['cell'])
        dbentry = (f'<{resi_name}> \n{comments} \nRESI {resi_name} \n'
                   f'{head} \n{cell} \n{atoms} \n</{resi_name}>\n')
        return dbentry


if __name__ == '__main__':
    """
    import doctest
    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))
    else:
        print('{} of {} tests failed'.format(failed, attempted))
    #homedir = expanduser("~")
    #userdb_path = os.path.join(homedir, "dsr_db.txt")
    """


    def std_dev_test(fragment, restraints, atoms, atnames, cell):
        for num, line in enumerate(restraints):
            prefixes = []
            dev = 0.02
            line = line.split()
            if not line:
                continue
            if line[0].upper() == 'SADI':
                prefixes.append(line[0])
                del line[0]
                try:
                    if not str(line[0][0]).isalpha():
                        prefixes.append(line[0])
                        dev = line[0]
                        del line[0]  # delete standard deviation
                except IndexError:
                    return False
                if len(line) % 2 == 1:  # test for uneven atoms count
                    print('*** Inconsistent SADI restraint line {} of "{}". '
                          'Not all atoms form a pair ***'.format(num, fragment))
                pairs = pairwise(line)
                distances = []
                pairlist = []
                if len(pairs) <= 2:
                    return True
                for i in pairs:
                    pairlist.append(i)
                    try:
                        atom1 = [float(x) for x in atoms[atnames.index(i[0])][2:5]]
                        atom2 = [float(x) for x in atoms[atnames.index(i[1])][2:5]]
                    except ValueError:
                        return False
                    dist = atomic_distance(atom1, atom2, cell)
                    distances.append(dist)
                if len(pairlist) == 2:
                    pairdev = 1 - (min(distances) / max(distances))
                    if pairdev > 2 * dev:
                        print('foooooooo', pairdev)
                stdev = std_dev(distances)
                print("esd < 0.065 ?: {:<4.4f}, esd < 2.5*sigma?: {:<4.4f} < {:<4.4f} -> {}".format(
                    stdev, stdev, 2.5 * float(dev), ("yes" if stdev < 2.5 * float(dev) else "no")))


    frag = 'WBVNT'

    dbpath = os.path.abspath('./src/dsr_shelx/dsr_db.txt')
    userpath = os.path.abspath('c:/Users/daniel/dsr_user_db.txt')
    db = ParseDB(dbpath, userdb_path=userpath)
    # pprint(db.databases['toluene'])
    db.check_consistency(frag)
    db.check_db_atom_consistency(frag)
    db.check_db_restraints_consistency(frag)
    db.check_sadi_consistence(frag)
    atnames = db.get_atomnames(frag)
    restr = db.get_restraints(frag)
    atoms = db.get_atoms(frag)
    cell = db.get_cell(frag)
    # print(db.databases['toluene'])
    print(dbpath)

    std_dev_test(fragment=frag, restraints=restr, atnames=atnames, atoms=atoms, cell=cell)



# -*- encoding: utf-8 -*-
# mÃ¶p
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import fnmatch
import os
import sys

import misc
from constants import RESTRAINT_CARDS



def add_residue_to_dfix(dfix_head, resinum):
    """
    Add a residue to a list of DFIX/DANG restraints
    DFIX 1.234 C1 C2 -> DFIX 1.234 C1_4 C2_4
    >>> add_residue_to_dfix(['DFIX 1.456 C1 C2', 'DFIX 1.212 C3 C4'], 4)
    ['DFIX  1.456  C1_4  C2_4\\n', 'DFIX  1.212  C3_4  C4_4\\n']
    >>> add_residue_to_dfix(['DFIX 1.456 C1 C2', 'DFIX 1.212 C3 C4'], '5')
    ['DFIX  1.456  C1_5  C2_5\\n', 'DFIX  1.212  C3_5  C4_5\\n']
    >>> add_residue_to_dfix(['FLAT C6 C1 C2 C3', 'FLAT  C7  C1  C2  C3'], '2')
    ['FLAT  C6_2  C1_2  C2_2  C3_2\\n', 'FLAT  C7_2  C1_2  C2_2  C3_2\\n']
    """
    newhead = []
    for line in dfix_head:
        line = line.split()
        first = line[0]
        for num, item in enumerate(line):
            try:
                int(item[0])
            except:
                line[num] = line[num] + '_' + str(resinum)
                continue
        line = first + '  ' + '  '.join(line[1:]) + '\n'
        newhead.append(line)
    return newhead


def collect_all_restraints(reslist):
    """
    :return all_restraints: list
    collects all restraints in the resfile and returns a list with them
    [['RIGU_CF3', 'O1', '>', 'F9'], '...']
    """
    all_restraints = []
    for n, resline in enumerate(reslist):
        resline = resline.strip(' \n\r')
        # resline = resline.split()
        try:
            resline[:4]
        except:
            continue
        if resline[:4] in RESTRAINT_CARDS:
            # see for the next  lines if the lines continues with "=":
            line = 0
            while resline[-1] == '=':
                resline = resline[:-1] + reslist[n + line + 1]
                line += 1
                if not resline[-1] == '=':
                    break
                if line > 500:
                    break
            all_restraints.append(resline)
    return all_restraints


def remove_duplicate_restraints(reslist, restraints, residue_class=''):
    """
    removes restraints from the header which are already
    in the res-file.

    :param restraints:         database header (list of strings)
    :param residue_class:  SHELXL residue class
    :type residue_class:   string
    :param all_restraints: all restraints in the res file
    :type all_restraints:  list
    :return new_restr: list
    >>> dbhead = ["SADI 0.02 C1 C2 C2 C3 C3 C4", "SADI 0.04 C1 C3 C3 C5", "DFIX 1.45 C1 C2"]
    >>> all_restraints = ["SADI C1 C2 C2 C3 C3 C4", "SADI C1 C3 C3 C5", "DFIX C1 C2", "SADI C4 C5 C5 C6"]
    >>> remove_duplicate_restraints(all_restraints, dbhead)
    <BLANKLINE>
    Already existing restraints were not applied again.
    []

    >>> all_restraints = ["SADI 0.02 C1 C2 C2 C3 C3 C4", "SADI 0.04 C1 C3 C3 C5", "DFIX 1.45 C1 C2"]
    >>> dbhead = ["SADI C1 C2 C2 C3 C3 C4", "SADI C1 C3 C3 C5", "DFIX C1 C2", "SADI C4 C5 C5 C6"]
    >>> remove_duplicate_restraints(all_restraints, dbhead)
    <BLANKLINE>
    Already existing restraints were not applied again.
    ['SADI C4 C5 C5 C6']
    """
    modified = False
    new_restr = restraints[:]
    for num, line in enumerate(restraints):
        line = remove_stddev_from_restraint(line.split())
        for restr in collect_all_restraints(reslist):
            restr = restr.split()
            restr = remove_stddev_from_restraint(restr)
            if line == restr:
                new_restr[num] = ''
                modified = True
                break
    if modified:
        if residue_class:
            print('\nAlready existing restraints for residue "{}" were not '
                  'applied again.'.format(residue_class))
        else:
            print('\nAlready existing restraints were not applied again.')
    new_restr = [x for x in new_restr if x]
    return new_restr


def remove_stddev_from_restraint(restr):
    # type: (list) -> list
    """
    Parameters
    ----------
    restr: list of restraints

    Returns list of restraints without standars deviation
    -------

    >>> r = ['SADI', '0.02', 'C1', 'C2', 'C3', 'C4']
    >>> r2 = ['SADI', 'C1', 'C2', 'C3', 'C4']
    >>> remove_stddev_from_restraint(r)
    ['SADI', 'C1', 'C2', 'C3', 'C4']
    >>> remove_stddev_from_restraint(r2)
    ['SADI', 'C1', 'C2', 'C3', 'C4']
    """
    new = []
    # find out where the atoms begin (leave out numbers):
    for num, i in enumerate(restr):
        if i[0].isalpha():
            new.append(i)
    return new


def write_dbhead_to_file(dsrp, filename, dbhead, resi_class, resi_number):
    """
    write the restraints to an external file
    :param filename:     filename of database file
    :param dbhead:       database header
    :param resi_class:   SHELXL residue class
    :param resi_number:  SHELXL residue number
    :return filename:    full file name where restraints will be written
    """
    number = '1'
    files = []
    # find a unique number for the restraint file:
    for filen in misc.sortedlistdir('.'):
        if fnmatch.fnmatch(filen, 'dsr_*_' + filename):
            filenum = filen.split('_')
            if str.isdigit(filenum[1]):
                files.append(filenum[1])
    filepath, filename = os.path.split(os.path.abspath(filename))
    try:
        number = str(int(files[-1]) + 1)
    except IndexError:
        pass
    if not dsrp.resiflag:  # no residues
        filename = 'dsr_' + number + '_' + filename
    if dsrp.resiflag and resi_number:  # only residue number known
        filename = 'dsr_' + resi_class + '_' + resi_number + '_' + filename
    if dsrp.resiflag and not resi_number and resi_class:  # only residue class known
        filename = 'dsr_' + resi_class + '_' + filename
    if os.path.isfile(os.path.abspath(filename)):
        print('Previous restraint file found. Using restraints from "{}"'.format(filename))
        return filename
    try:
        dfix_file = open(os.path.join(filepath, filename), 'w')  # open the ins file
    except IOError:
        print('*** Unable to write restraints file! Check directory write permissions. ***')
        sys.exit(False)
    print('Restraints were written to "{}"'.format(os.path.join(filepath, filename)))
    for i in dbhead:  # modified reslist
        dfix_file.write("%s" % i)  # write the new file
    dfix_file.close()
    return filename


if __name__ == '__main__':
    import doctest

    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))



"""
This file is used to export the DSR database to sqlite for Olex2.
Every fragemnt needs a picture with the same name as the fragment tag in ./pictures
"""
import os
import sqlite3 as lite
import sys

import misc
from constants import SHX_CARDS
from dbfile import ParseDB

misc.remove_file('./fragment-database.sqlite')
gl = ParseDB()
db = gl.parse(dbpath="./olex_dsr_db.txt", dbname='olex_db')

dbfilename = 'fragment-database.sqlite'
con = lite.connect(dbfilename)
con.text_factory = str
cur = con.cursor()


def get_fragment_atoms_cartesian(fragment):
    """
    returns the coordinates of the fragment as cartesian coords
    as list of lists [['-2.7538', '15.9724', '22.6810'], ['0.7939', '16.3333', '21.3135'], ...
    :param fragment:
    :type fragment:
    """
    from export import Export
    ex = Export(gl, invert=False)
    atoms = ex.format_atoms_for_export(gl.get_cell(fragment), gl.get_atoms(fragment), False)
    coords = []
    for i in atoms:
        co = i.split()[2:5]
        co = ['{:>.4f}'.format(float(x)) for x in co]
        # print(co)
        coords.append(co)
    return coords


def initialize_db(con, cur):
    print('beginning export')
    con.execute("PRAGMA foreign_keys = ON")
    cur.execute("DROP TABLE IF EXISTS fragment")
    cur.execute("DROP TABLE IF EXISTS atoms")
    cur.execute("DROP TABLE IF EXISTS atom")
    cur.execute("DROP TABLE IF EXISTS FragmentRestraints")
    cur.execute("DROP TABLE IF EXISTS Restraints")
    try:
        cur.execute("DROP INDEX Atoms_FK")
    except:
        pass
    try:
        cur.execute("DROP INDEX Restraint_FK")
    except:
        pass
    try:
        cur.execute("DROP INDEX Fragment_Name")
    except:
        pass
    try:
        cur.execute("DROP INDEX AtomId")
    except:
        pass

    cur.execute('''
                CREATE TABLE Fragment (
                    Id    INTEGER NOT NULL,
                    class  VARCHAR(4),
                    version TEXT,
                    Name    TEXT,
                    Reference    TEXT,
                    comment    TEXT,
                    picture    BLOB,
                    PRIMARY KEY(Id));
                ''')
    cur.execute('''
                CREATE TABLE Atoms (
                    Id    INTEGER NOT NULL,
                    FragmentId    INTEGER NOT NULL,
                    version TEXT,
                    Name    VARCHAR(255),
                    element    VARCHAR(2),
                    x    FLOAT,
                    y    FLOAT,
                    z    FLOAT,
                PRIMARY KEY(Id),
                  FOREIGN KEY(FragmentId)
                    REFERENCES Fragment(Id)
                      ON DELETE CASCADE
                      ON UPDATE NO ACTION);
                ''')
    cur.execute('''
                   CREATE TABLE Restraints (
                      Id INTEGER  NOT NULL,
                      FragmentId  INTEGER NOT NULL,
                      version TEXT,
                      ShelxName CHAR(4),
                      Atoms TEXT,
                    PRIMARY KEY(Id),
                      FOREIGN KEY(FragmentId)
                        REFERENCES Fragment(Id)
                        ON DELETE CASCADE
                        ON UPDATE NO ACTION);
                    ''')

    cur.execute('''
                CREATE INDEX Atoms_FK ON Atoms(FragmentId);
                ''')
    cur.execute('''
                CREATE INDEX Restraint_FK ON Restraints(FragmentId);
                ''')
    cur.execute('''
                CREATE INDEX Fragment_Name ON Atoms(Name);
                ''')
    cur.execute('''
                CREATE INDEX AtomId ON Atoms(Id);
                ''')
    con.execute("PRAGMA foreign_keys = ON")


def fill_fragment_table(table):
    cur.execute('''INSERT INTO fragment (class, version, name, reference, comment, picture)
                                VALUES(?, ?, ?, ?, ?, ?)''', table)


def fill_restraints_table(fragmentid, head):
    for line in head:
        if not line:
            continue
        restr_table = []
        if line[:4] in SHX_CARDS:
            restr_table.append(fragmentid)
            restr_table.append('1')
            restr_table.append(line[:4])
            restr_table.append(line[5:])
        cur.execute('''INSERT INTO Restraints (fragmentid, version, ShelxName, atoms)
                                VALUES(?, ?, ?, ?)''', restr_table)


def fill_atoms(FragmentId, fragment):
    atom_names = gl.get_atomnames(fragment)
    coords = get_fragment_atoms_cartesian(fragment)
    aelements = gl.get_atomic_numbers(fragment)
    for atname, coord, element in zip(atom_names, coords, aelements):
        x = coord[0]
        y = coord[1]
        z = coord[2]
        table_atoms = (FragmentId, '1', atname, element, x, y, z)
        cur.execute('''INSERT INTO atoms (FragmentId, version, Name, element,
                        x, y, z) VALUES(?, ?, ?, ?, ?, ?, ?)''', table_atoms)


def get_picture(name):
    try:
        fexist = os.stat('./pictures/{}.png'.format(name))
    except OSError:
        fexist = False
    if fexist:
        with open('./pictures/{}.png'.format(name), 'rb') as f:
            return f.read()
    else:
        print('   #####  No picture  for {} found!!!'.format(name))
        sys.exit()


def export_database():
    for fid, fragment in enumerate(list(db.keys()), 1):
        Name = gl.get_fragment_name(fragment)
        print("Exporting {}: {}: {}".format(fid, fragment, Name))
        head = db[fragment]['restraints']
        name = db[fragment]['name']
        # formula = db.get_sum_formula(fragment)
        reference = gl.get_src(fragment)
        resiclass = gl.get_resi(fragment)
        picture = get_picture(fragment)
        picture = lite.Binary(picture)
        table_frag = (resiclass, '1', Name, reference, name, picture)
        fill_fragment_table(table_frag)
        fill_atoms(fid, fragment)
        fill_restraints_table(fid, head)
    print('\nDatabase exported to "{}"'.format(dbfilename))

    rows = cur.fetchall()
    for row in rows:
        print("{}  {}  {:>8.4f} {:>8.4f} {:>8.4f}".format(*row))
        # print("{}".format(row))


# enable to re-export the db to sql:
initialize_db(con=con, cur=cur)
# Create empty user database:
# initialize_db(con=conu, cur=curu)
export_database()
con.commit()

# t1 = time.clock()
# print_table(con)
# t2 = time.clock()
# print(t2-t1)
